#!/usr/bin/env python
# -*- coding: utf-8 -*-

# The MIT License (MIT)
# Copyright (c) 2020 Daniel Schick
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
# DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
# OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE
# OR OTHER DEALINGS IN THE SOFTWARE.

__all__ = ['Layer', 'AmorphousLayer', 'UnitCell']

__docformat__ = 'restructuredtext'

from .atoms import Atom, AtomMixed
from .. import u, Q_
import numpy as np
from inspect import isfunction
from sympy import integrate, lambdify, symbols, symarray
from tabulate import tabulate
import scipy.constants as constants
import warnings


class Layer:
    r"""Layer

    Base class of real physical layers, such as amorphous layers and unit cells.

    It holds different structural, thermal, and elastic properties that are
    relevant for simulations.

    Args:
        id (str): id of the layer.
        name (str): name of the layer.

    Keyword Args:
        roughness (float): gaussian width of the top roughness of a layer.
        deb_wal_fac (float): Debye Waller factor.
        sound_vel (float): sound velocity.
        phonon_damping (float): phonon damping.
        opt_pen_depth (float): optical penetration depth.
        opt_ref_index (float): refractive index.
        opt_ref_index_per_strain (float): change of refractive index per strain.
        heat_capacity (float): heat capacity.
        therm_cond (float): thermal conductivity.
        lin_therm_exp (float): linear thermal expansion.
        sub_system_coupling (float): sub-system coupling.

    Attributes:
        id (str): id of the layer.
        name (str): name of the layer.
        thickness (float): thickness of the layer [m].
        mass (float): mass of the layer [kg].
        mass_unit_area (float): mass of layer normalized to unit area of 1 Å² [kg].
        density (float): density of the layer [kg/m³].
        area (float): area of layer [m²].
        volume (float): volume of layer [m³].
        roughness (float): gaussian width of the top roughness of a layer [m].
        deb_wal_fac (list[@lambda]): list of T-dependent Debye-Waller factors
            `\langle u^2\rangle` [m²].
        sound_vel (float): longitudinal sound velocity in the layer [m/s].
        spring_const (ndarray[float]): spring constant of the layer [kg/s²]
            and higher orders.
        phonon_damping (float): damping constant of phonon propagation [kg/s].
        opt_pen_depth (float): optical penetration depth of the layer [m].
        opt_ref_index (ndarray[float]): optical refractive index - real
           and imagenary part :math:`n + i\kappa`.
        opt_ref_index_per_strain (ndarray[float]): optical refractive
           index change per strain - real and imagenary part
           :math:`\frac{d n}{d \eta} + i\frac{d \kappa}{d \eta}`.
        therm_cond (list[@lambda]): list of T-dependent thermal conductivity
            [W/(m K)].
        lin_therm_exp (list[@lambda]): list of T-dependent linear thermal
           expansion coefficient (relative).
        int_lin_therm_exp (list[@lambda]): list of T-dependent integrated
           linear thermal expansion coefficient.
        heat_capacity (list[@lambda]): list of T-dependent heat capacity
           function [J/(kg K)].
        int_heat_capacity (list[@lambda]): list of T-dependent integrated heat
           capacity function.
        sub_system_coupling (list[@lambda]): list of coupling functions of
           different subsystems [W/m³].
        num_sub_systems (int): number of subsystems for heat and phonons
           (electrons, lattice, spins, ...).
        eff_spin (float): effective spin.
        curie_temp (float): Curie temperature [K].
        mf_exch_coupling (float): mean field exchange coupling constant [m²kg/s²].
        lamda (float): intrinsic coupling to bath parameter.
        mag_moment (float): atomic magnetic moment [mu_Bohr].
        aniso_exponent(ndarray[float]): exponent of T-dependence uniaxial
            anisotropy.
        anisotropy (ndarray[float]): anisotropy at T=0 K [J/m³] as x,y,z component vector.
        exch_stiffness (float): exchange stiffness at T=0 K [J/m].
        mag_saturation (float): saturation magnetization at 0 K [J/T/m³].

    """

    def __init__(self, id, name, **kwargs):
        self.id = id
        self.name = name
        self.num_sub_systems = 1
        self.roughness = kwargs.get('roughness', 0.0*u.nm)
        self.spring_const = np.array([0.0])
        self.deb_wal_fac = kwargs.get('deb_wal_fac', 0.0)
        self.sound_vel = kwargs.get('sound_vel', 0.0*u.m/u.s)
        self.phonon_damping = kwargs.get('phonon_damping', 0.0*u.kg/u.s)
        self.opt_pen_depth = kwargs.get('opt_pen_depth', 0.0*u.nm)
        self.opt_ref_index = kwargs.get('opt_ref_index', 0.0+0.0j)
        self.opt_ref_index_per_strain = kwargs.get('opt_ref_index_per_strain', 0.0+0.0j)
        self.heat_capacity = kwargs.get('heat_capacity', 0.0)
        self.therm_cond = kwargs.get('therm_cond', 0.0)
        self.lin_therm_exp = kwargs.get('lin_therm_exp', 0.0)
        self.sub_system_coupling = kwargs.get('sub_system_coupling', 0.0)

        if len(self.heat_capacity) == len(self.therm_cond) \
                == len(self.lin_therm_exp) == len(self.sub_system_coupling):
            self.num_sub_systems = len(self.heat_capacity)
        else:
            raise ValueError('Heat capacity, thermal conductivity, linear '
                             'thermal expansion, and subsystem coupling have not '
                             'the same number of elements!')

        self.eff_spin = kwargs.get('eff_spin', 0.0)
        self.curie_temp = kwargs.get('curie_temp', 0.0*u.K)
        self.lamda = kwargs.get('lamda', 0.0)
        self.mag_moment = kwargs.get('mag_moment', 0.0*u.bohr_magneton)
        self.aniso_exponent = kwargs.get('aniso_exponent', 0.0)
        self.anisotropy = kwargs.get('anisotropy', [0.0, 0.0, 0.0]*u.J/u.m**3)
        self.exch_stiffness = kwargs.get('exch_stiffness', 0.0*u.J/u.m)
        self.mag_saturation = kwargs.get('mag_saturation', 0.0*u.J/u.T/u.m**3)

    def __str__(self):
        """String representation of this class"""
        output = []
        try:
            output += [['area', '{:.4g~P}'.format(self.area.to('nm**2'))],
                       ['volume', '{:.4g~P}'.format(self.volume.to('nm**3'))]]
        except AttributeError:
            output += [['no area or volume set', '']]
        try:
            output += [['mass', '{:.4g~P}'.format(self.mass.to('kg'))],
                       ['mass per unit area', '{:.4g~P}'.format(self.mass_unit_area)],
                       ['density', '{:.4g~P}'.format(self.density.to('kg/meter**3'))]]
        except AttributeError:
            output += [['no mass set', '']]
        output += [['roughness', '{:.4g~P}'.format(self.roughness.to('nm'))],
                   ['Debye Waller factor', ' m²\n'.join(self.deb_wal_fac_str) + ' m²'],
                   ['sound velocity', '{:.4g~P}'.format(self.sound_vel.to('meter/s'))],
                   ['spring constant', '{:.4g~P}'.format(self.spring_const * u.kg/u.s**2)],
                   ['phonon damping', '{:.4g~P}'.format(self.phonon_damping.to('kg/s'))],
                   ['opt. pen. depth', '{:.4g~P}'.format(self.opt_pen_depth.to('nm'))],
                   ['opt. refractive index', '{0.real:.4f} + {0.imag:.4f}i'.format(
                       self.opt_ref_index)],
                   ['opt. ref. index/strain', '{0.real:.4f} + {0.imag:.4f}i'.format(
                       self.opt_ref_index_per_strain)],
                   ['thermal conduct.', ' W/(m K)\n'.join(self.therm_cond_str) + ' W/(m K)'],
                   ['linear thermal expansion', '\n'.join(self.lin_therm_exp_str)],
                   ['heat capacity', ' J/(kg K)\n'.join(self.heat_capacity_str) + ' J/(kg K)'],
                   ['subsystem coupling', ' W/m³\n'.join(self.sub_system_coupling_str) + ' W/m³'],
                   ['effective spin', self.eff_spin],
                   ['Curie temperature', '{:.4g~P}'.format(self.curie_temp.to('K'))],
                   ['mean-field exch. coupling', '{:.4g~P}'.format(
                       self.mf_exch_coupling.to('m**2*kg/s**2'))],
                   ['coupling to bath parameter', self.lamda],
                   ['atomic magnetic moment', '{:.4g~P}'.format(self.mag_moment.to(
                       'bohr_magneton'))],
                   ['uniaxial anisotropy exponent', self.aniso_exponent],
                   ['anisotropy', '{:.4g~P}'.format(self.anisotropy.to('J/m**3'))],
                   ['exchange stiffness', '{:.4g~P}'.format(self.exch_stiffness.to('J/m'))],
                   ['saturation magnetization', '{:.4g~P}'.format(
                       self.mag_saturation.to('J/T/m**3'))]]

        return output

    def check_input(self, inputs, change_num_sub_systems=True):
        """check_input

        Checks the input and create a list of function handle strings with T as
        argument. Inputs can be strings, floats, ints, or pint quantities.

        Args:
            inputs (list[str, int, float, Quantity]): list of strings, int, floats,
                or Pint quantities.
            change_num_sub_systems (boolean, optional): wheather the number of
                sub-systems should be changed. Defaults to True.

        Returns:
            (tuple):
            - *output (list[@lambda])* - list of lambda functions.
            - *output_strs (list[str])* - list of string-representations.

        """
        output = []
        output_strs = []
        # if the input is not a list, we convert it to one
        if not isinstance(inputs, list):
            inputs = [inputs]
        # update number of subsystems
        K = self.num_sub_systems
        k = len(inputs)
        if k != K and change_num_sub_systems:
            print('Number of subsystems changed from {:d} to {:d}.'.format(K, k))
            self.num_sub_systems = k

        # traverse each list element and convert it to a function handle
        for input in inputs:
            T = symbols('T')
            if isfunction(input):
                raise ValueError('Please use string representation of function!')
            elif isinstance(input, str):
                try:
                    # backwards compatibility for direct lambda definition
                    if ':' in input:
                        # strip lambda prefix
                        input = input.split(':')[1]
                    # backwards compatibility for []-indexing
                    input = input.replace('[', '_').replace(']', '')
                    # check for presence of indexing and use symarray as argument
                    if '_' in input:
                        T = symarray('T', k)
                        output.append(lambdify([T], input, modules='numpy'))
                    else:
                        output.append(lambdify(T, input, modules='numpy'))
                    output_strs.append(input.strip())
                except Exception as e:
                    print('String input for layer property ' + input + ' \
                        cannot be converted to function handle!')
                    print(e)
            elif isinstance(input, (int, float)):
                output.append(lambdify(T, input, modules='numpy'))
                output_strs.append(str(float(input)))
            elif isinstance(input, object):
                output.append(lambdify(T, input.to_base_units().magnitude, modules='numpy'))
                output_strs.append(str(float(input.to_base_units().magnitude)))
            else:
                raise ValueError('Layer property input has to be a single or '
                                 'list of numerics, Quantities, or function handle strings '
                                 'which can be converted into a lambda function!')

        return output, output_strs

    def get_property_dict(self, **kwargs):
        """get_property_dict

        Returns a dictionary with all parameters. objects or dicts and
        objects are converted to strings. if a type is given, only these
        properties are returned.

        Args:
            **kwargs (list[str]): types of requested properties.

        Returns:
            R (dict): dictionary with requested properties.

        """
        # initialize input parser and define defaults and validators
        properties_by_types = {'heat': ['_thickness', '_mass_unit_area', '_density',
                                        '_opt_pen_depth', 'opt_ref_index',
                                        'therm_cond_str', 'heat_capacity_str',
                                        'int_heat_capacity_str', 'sub_system_coupling_str',
                                        'num_sub_systems'],
                               'phonon': ['num_sub_systems', 'int_lin_therm_exp_str', '_thickness',
                                          '_mass_unit_area', 'spring_const', '_phonon_damping'],
                               'xray': ['num_atoms', '_area', '_mass', 'deb_wal_fac_str',
                                        '_thickness'],
                               'optical': ['_c_axis', '_opt_pen_depth', 'opt_ref_index',
                                           'opt_ref_index_per_strain'],
                               'magnetic': ['_thickness', 'magnetization', 'eff_spin',
                                            '_curie_temp', '_aniso_exponents', '_anisotropy',
                                            '_exch_stiffness', '_mag_saturation', 'lamda'],
                               }

        types = (kwargs.get('types', 'all'))
        if not type(types) is list:
            types = [types]
        attrs = vars(self)
        R = {}
        for t in types:
            # define the property names by the given type
            if t == 'all':
                return attrs
            else:
                S = dict((key, value) for key, value in attrs.items()
                         if key in properties_by_types[t])
                R.update(S)

        return R

    def get_acoustic_impedance(self):
        """get_acoustic_impedance

        Calculates the acoustic impedance.

        Returns:
            Z (float): acoustic impedance.

        """
        Z = np.sqrt(self.spring_const[0] * self.mass/self.area**2)
        return Z

    def set_ho_spring_constants(self, HO):
        """set_ho_spring_constants

        Set the higher orders of the spring constant for anharmonic
        phonon simulations.

        Args:
            HO (ndarray[float]): higher order spring constants.

        """
        # reset old higher order spring constants
        self.spring_const = np.delete(self.spring_const, np.r_[1:len(self.spring_const)])
        self.spring_const = np.hstack((self.spring_const, HO))

    def set_opt_pen_depth_from_ref_index(self, wavelength):
        """set_opt_pen_depth_from_ref_index

        Set the optical penetration depth from the optical referactive index
        for a given wavelength.

        Args:
            wavelength (Quantity): wavelength as Pint Quantitiy.

        """
        if np.imag(self.opt_ref_index) == 0:
            self.opt_pen_depth = Q_(np.inf, u.m)
        else:
            self.opt_pen_depth = wavelength/(4*np.pi*np.abs(np.imag(self.opt_ref_index)))

    def calc_spring_const(self):
        r"""calc_spring_const

        Calculates the spring constant of the layer from the mass per unit area,
        sound velocity and thickness

        .. math:: k = m \, \left(\frac{v}{c}\right)^2

        """
        try:
            self.spring_const[0] = (self._mass_unit_area * (self._sound_vel/self._thickness)**2)
        except AttributeError:
            # no mass set, yet
            self.spring_const[0] = 0

    def calc_mf_exchange_coupling(self):
        r"""calc_mf_exchange_coupling

        Calculate the mean-field exchange coupling constant

        .. math:: J = \frac{3}{S_{eff}+1} k_B T_C

        """
        try:
            self._mf_exch_coupling = 3*self.eff_spin/(self.eff_spin+1)*constants.k*self._curie_temp
        except AttributeError:
            # on initialization self._curie_temp
            self._mf_exch_coupling = 0

    @property
    def thickness(self):
        return Q_(self._thickness, u.meter).to('nm')

    @thickness.setter
    def thickness(self, thickness):
        self._thickness = thickness.to_base_units().magnitude

    @property
    def mass(self):
        return Q_(self._mass, u.kg)

    @mass.setter
    def mass(self, mass):
        self._mass = mass.to_base_units().magnitude

    @property
    def mass_unit_area(self):
        return Q_(self._mass_unit_area, u.kg)

    @mass_unit_area.setter
    def mass_unit_area(self, mass_unit_area):
        self._mass_unit_area = mass_unit_area.to_base_units().magnitude

    @property
    def density(self):
        return Q_(self._density, u.kg/u.m**3)

    @density.setter
    def density(self, density):
        self._density = density.to_base_units().magnitude

    @property
    def area(self):
        return Q_(self._area, u.m**2)

    @area.setter
    def area(self, area):
        self._area = area.to_base_units().magnitude

    @property
    def volume(self):
        return Q_(self._volume, u.m**3)

    @volume.setter
    def volume(self, volume):
        self._volume = volume.to_base_units().magnitude

    @property
    def deb_wal_fac(self):
        return self._deb_wal_fac

    @deb_wal_fac.setter
    def deb_wal_fac(self, deb_wal_fac):
        self._deb_wal_fac, self.deb_wal_fac_str = self.check_input(deb_wal_fac, False)

    @property
    def sound_vel(self):
        return Q_(self._sound_vel, u.m/u.s)

    @sound_vel.setter
    def sound_vel(self, sound_vel):
        # spring constants are (re)calculated on setting the sound velocity
        self._sound_vel = sound_vel.to_base_units().magnitude
        self.calc_spring_const()

    @property
    def phonon_damping(self):
        return Q_(self._phonon_damping, u.kg/u.s)

    @phonon_damping.setter
    def phonon_damping(self, phonon_damping):
        self._phonon_damping = phonon_damping.to_base_units().magnitude

    @property
    def opt_pen_depth(self):
        return Q_(self._opt_pen_depth, u.meter).to('nanometer')

    @opt_pen_depth.setter
    def opt_pen_depth(self, opt_pen_depth):
        self._opt_pen_depth = opt_pen_depth.to_base_units().magnitude

    @property
    def roughness(self):
        return Q_(self._roughness, u.meter).to('nm')

    @roughness.setter
    def roughness(self, roughness):
        self._roughness = roughness.to_base_units().magnitude

    @property
    def heat_capacity(self):
        return self._heat_capacity

    @heat_capacity.setter
    def heat_capacity(self, heat_capacity):
        # (re)calculate the integrated heat capacity
        self._heat_capacity, self.heat_capacity_str = self.check_input(heat_capacity)
        # delete last anti-derivative
        self._int_heat_capacity = None
        # recalculate the anti-derivative
        self.int_heat_capacity

    @property
    def therm_cond(self):
        return self._therm_cond

    @therm_cond.setter
    def therm_cond(self, therm_cond):
        self._therm_cond, self.therm_cond_str = self.check_input(therm_cond)

    @property
    def int_heat_capacity(self):
        if hasattr(self, '_int_heat_capacity') and isinstance(self._int_heat_capacity, list):
            return self._int_heat_capacity
        else:
            self._int_heat_capacity = []
            self.int_heat_capacity_str = []
            T = symbols('T')
            try:
                for hcs in self.heat_capacity_str:
                    integral = integrate(hcs, T)
                    self._int_heat_capacity.append(lambdify(T, integral, modules='numpy'))
                    self.int_heat_capacity_str.append(str(integral))
            except Exception as e:
                print('The sympy integration did not work. You can set the '
                      'analytical anti-derivative of the heat capacity '
                      'of your layer as function str of the temperature '
                      'T by typing layer.int_heat_capacity = \'c(T)\' '
                      'where layer is the name of the layer object.')
                print(e)

        return self._int_heat_capacity

    @int_heat_capacity.setter
    def int_heat_capacity(self, int_heat_capacity):
        self._int_heat_capacity, self.int_heat_capacity_str = self.check_input(
                int_heat_capacity)

    @property
    def lin_therm_exp(self):
        return self._lin_therm_exp

    @lin_therm_exp.setter
    def lin_therm_exp(self, lin_therm_exp):
        # (re)calculate the integrated linear thermal expansion coefficient
        self._lin_therm_exp, self.lin_therm_exp_str = self.check_input(lin_therm_exp)
        # delete last anti-derivative
        self._int_lin_therm_exp = None
        # recalculate the anti-derivative
        self.int_lin_therm_exp

    @property
    def int_lin_therm_exp(self):
        if hasattr(self, '_int_lin_therm_exp') and isinstance(self._int_lin_therm_exp, list):
            return self._int_lin_therm_exp
        else:
            self._int_lin_therm_exp = []
            self.int_lin_therm_exp_str = []
            T = symbols('T')
            try:
                for ltes in self.lin_therm_exp_str:
                    integral = integrate(ltes, T)
                    self._int_lin_therm_exp.append(lambdify(T, integral, modules='numpy'))
                    self.int_lin_therm_exp_str.append(str(integral))
            except Exception as e:
                print('The sympy integration did not work. You can set the '
                      'analytical anti-derivative of the linear thermal expansion '
                      'of your unit cells as lambda function of the temperature '
                      'T by typing layer.int_lin_therm_exp = \'c(T)\' '
                      'where layer is the name of the layer object.')
                print(e)

        return self._int_lin_therm_exp

    @int_lin_therm_exp.setter
    def int_lin_therm_exp(self, int_lin_therm_exp):
        self._int_lin_therm_exp, self.int_lin_therm_exp_str = self.check_input(
                int_lin_therm_exp)

    @property
    def sub_system_coupling(self):
        return self._sub_system_coupling

    @sub_system_coupling.setter
    def sub_system_coupling(self, sub_system_coupling):
        self._sub_system_coupling, self.sub_system_coupling_str = \
            self.check_input(sub_system_coupling)

    @property
    def eff_spin(self):
        return self._eff_spin

    @eff_spin.setter
    def eff_spin(self, eff_spin):
        self._eff_spin = float(eff_spin)
        self.calc_mf_exchange_coupling()

    @property
    def curie_temp(self):
        return Q_(self._curie_temp, u.K)

    @property
    def mf_exch_coupling(self):
        return Q_(self._mf_exch_coupling, u.m**2*u.kg/(u.s**2))

    @curie_temp.setter
    def curie_temp(self, curie_temp):
        self._curie_temp = float(curie_temp.to_base_units().magnitude)
        self.calc_mf_exchange_coupling()

    @property
    def mag_moment(self):
        return Q_(self._mag_moment, u.A*u.m**2).to('bohr_magneton')

    @mag_moment.setter
    def mag_moment(self, mag_moment):
        self._mag_moment = float(mag_moment.to_base_units().magnitude)

    @property
    def anisotropy(self):
        return Q_(self._anisotropy, u.J/u.m**3)

    @anisotropy.setter
    def anisotropy(self, anisotropy):
        self._anisotropy = np.zeros(3)
        try:
            if len(anisotropy) == 3:
                self._anisotropy = anisotropy.to_base_units().magnitude
            else:
                warnings.warn('Anisotropy must be a scalar or vector of length 3!')
        except TypeError:
            self._anisotropy[0] = anisotropy.to_base_units().magnitude

    @property
    def exch_stiffness(self):
        return Q_(self._exch_stiffness, u.J/u.m)

    @exch_stiffness.setter
    def exch_stiffness(self, exch_stiffness):
        self._exch_stiffness = np.zeros(3)
        try:
            if len(exch_stiffness) == 3:
                self._exch_stiffness = exch_stiffness.to_base_units().magnitude
            else:
                warnings.warn('Exchange stiffness must be a scalar or vector of length 3!')
        except TypeError:
            self._exch_stiffness[:] = exch_stiffness.to_base_units().magnitude

    @property
    def mag_saturation(self):
        return Q_(self._mag_saturation, u.J/u.T/u.m**3)

    @mag_saturation.setter
    def mag_saturation(self, mag_saturation):
        self._mag_saturation = float(mag_saturation.to_base_units().magnitude)


class AmorphousLayer(Layer):
    r"""AmorphousLayer

    Representation of amorphous layers containing an Atom or AtomMixed.

    Args:
        id (str): id of the layer.
        name (str): name of layer.
        thickness (float): thickness of the layer.
        density (float): density of the layer.

    Keyword Args:
        atom (object): Atom or AtomMixed in the layer.
        roughness (float): gaussian width of the top roughness of a layer.
        deb_wal_fac (float): Debye Waller factor.
        sound_vel (float): sound velocity.
        phonon_damping (float): phonon damping.
        roughness (float): gaussian width of the top roughness of a layer.
        opt_pen_depth (float): optical penetration depth.
        opt_ref_index (float): refractive index.
        opt_ref_index_per_strain (float): change of refractive index per
           strain.
        heat_capacity (float): heat capacity.
        therm_cond (float): thermal conductivity.
        lin_therm_exp (float): linear thermal expansion.
        sub_system_coupling (float): sub-system coupling.

    Attributes:
        id (str): id of the layer.
        name (str): name of the layer.
        thickness (float): thickness of the layer [m].
        mass (float): mass of the layer [kg].
        mass_unit_area (float): mass of layer normalized to unit area of 1 Å² [kg].
        density (float): density of the layer [kg/m³].
        area (float): area of layer [m²].
        volume (float): volume of layer [m³].
        roughness (float): gaussian width of the top roughness of a layer [m].
        deb_wal_fac (list[@lambda]): list of T-dependent Debye-Waller factors
            `\langle u^2\rangle` [m²].
        sound_vel (float): longitudinal sound velocity in the layer [m/s].
        spring_const (ndarray[float]): spring constant of the layer [kg/s²]
            and higher orders.
        phonon_damping (float): damping constant of phonon propagation [kg/s].
        opt_pen_depth (float): optical penetration depth of the layer [m].
        opt_ref_index (ndarray[float]): optical refractive index - real
           and imagenary part :math:`n + i\kappa`.
        opt_ref_index_per_strain (ndarray[float]): optical refractive
           index change per strain - real and imagenary part
           :math:`\frac{d n}{d \eta} + i\frac{d \kappa}{d \eta}`.
        therm_cond (list[@lambda]): list of HANDLES T-dependent thermal
           conductivity [W/(m K)].
        lin_therm_exp (list[@lambda]): list of T-dependent linear thermal
           expansion coefficient (relative).
        int_lin_therm_exp (list[@lambda]): list of T-dependent integrated
           linear thermal expansion coefficient.
        heat_capacity (list[@lambda]): list of T-dependent heat capacity
           function [J/(kg K)].
        int_heat_capacity (list[@lambda]): list of T-dependent integrated heat
           capacity function.
        sub_system_coupling (list[@lambda]): list of of coupling functions of
           different subsystems [W/m³].
        num_sub_systems (int): number of subsystems for heat and phonons
           (electrons, lattice, spins, ...).
        eff_spin (float): effective spin.
        curie_temp (float): Curie temperature [K].
        mf_exch_coupling (float): mean field exchange coupling constant [m²kg/s²].
        lamda (float): intrinsic coupling to bath parameter.
        mag_moment (float): atomic magnetic moment [mu_Bohr].
        aniso_exponent(ndarray[float]): exponent of T-dependence uniaxial
            anisotropy.
        anisotropy (float): anisotropy at T=0 K [J/m³] as x,y,z component vector.
        exch_stiffness (float): exchange stiffness at T=0 K [J/m].
        mag_saturation (float): saturation magnetization at 0 K [J/T/m³].
        magnetization (dict[float]): magnetization amplitude, phi and
           gamma angle inherited from the atom.
        atom (object): Atom or AtomMixed in the layer.

    """

    def __init__(self, id, name, thickness, density, **kwargs):
        self.thickness = thickness
        self.density = density
        self.area = 1.0*u.angstrom**2  # set as unit area
        self.volume = self.area*self.thickness
        self.mass = self.density*self.volume
        self.mass_unit_area = self.mass
        self.atom = kwargs.get('atom', [])
        super().__init__(id, name, **kwargs)

    def __str__(self):
        """String representation of this class"""
        output = [['id', self.id],
                  ['name', self.name],
                  ['thickness', '{:.4g~P}'.format(self.thickness)],
                  ]
        output += super().__str__()

        try:
            output += [['atom', self.atom.name],
                       ['magnetization', ''],
                       ['amplitude', self.magnetization['amplitude']],
                       ['phi [°]', '{:.4g~P}'.format(self.magnetization['phi'].to('deg'))],
                       ['gamma [°]', '{:.4g~P}'.format(self.magnetization['gamma'].to('deg'))], ]
        except AttributeError:
            output += [['no atom set', '']]

        class_str = 'Amorphous layer with the following properties\n\n'
        class_str += tabulate(output, headers=['parameter', 'value'], tablefmt='rst',
                              colalign=('right',), floatfmt=('.2f', '.2f'))
        return class_str

    @property
    def atom(self):
        return self._atom

    @atom.setter
    def atom(self, atom):
        if atom == []:  # no atom is set
            self.magnetization = {'amplitude': 0,
                                  'phi': 0*u.deg,
                                  'gamma': 0*u.deg,
                                  }
            return

        if not isinstance(atom, (Atom, AtomMixed)):
            raise ValueError('Class '
                             + type(atom).__name__
                             + ' is no possible atom of an amorphous layer. '
                             + 'Only Atom and AtomMixed are allowed!')
        self._atom = atom
        self.magnetization = {'amplitude': atom.mag_amplitude,
                              'phi': atom.mag_phi,
                              'gamma': atom.mag_gamma,
                              }

    @property
    def magnetization(self):
        return {'amplitude': self._magnetization['amplitude'],
                'phi': Q_(self._magnetization['phi'], u.rad).to('deg'),
                'gamma': Q_(self._magnetization['gamma'], u.rad).to('deg')
                }

    @magnetization.setter
    def magnetization(self, magnetization):
        self._magnetization = {'amplitude': magnetization['amplitude'],
                               'phi': magnetization['phi'].to_base_units().magnitude,
                               'gamma': magnetization['gamma'].to_base_units().magnitude
                               }


class UnitCell(Layer):
    r"""UnitCell

    Representation of unit cells made of one or multiple Atom or AtomMixed
    instances at defined positions.

    Args:
        id (str): id of the UnitCell.
        name (str): name of the UnitCell.
        c_axis (float): c-axis of the UnitCell.

    Keyword Args:
        a_axis (float): a-axis of the UnitCell.
        b_axis (float): b-axis of the UnitCell.
        deb_wal_fac (float): Debye Waller factor.
        sound_vel (float): sound velocity.
        phonon_damping (float): phonon damping.
        roughness (float): gaussian width of the top roughness of a layer.
        opt_pen_depth (float): optical penetration depth.
        opt_ref_index (float): refractive index.
        opt_ref_index_per_strain (float): change of refractive index per
           strain.
        heat_capacity (float): heat capacity.
        therm_cond (float): thermal conductivity.
        lin_therm_exp (float): linear thermal expansion.
        sub_system_coupling (float): sub-system coupling.

    Attributes:
        id (str): id of the layer.
        name (str): name of the layer.
        c_axis (float): out-of-plane c-axis [m].
        a_axis (float): in-plane a-axis [m].
        b_axis (float): in-plane b-axis [m].
        thickness (float): thickness of the layer [m].
        mass (float): mass of the layer [kg].
        mass_unit_area (float): mass of layer normalized to unit area of 1 Å² [kg].
        density (float): density of the layer [kg/m³].
        area (float): area of layer [m²].
        volume (float): volume of layer [m³].
        roughness (float): gaussian width of the top roughness of a layer [m].
        deb_wal_fac (list[@lambda]): list of T-dependent Debye-Waller factors
            `\langle u^2\rangle` [m²].
        sound_vel (float): longitudinal sound velocity in the layer [m/s].
        spring_const (ndarray[float]): spring constant of the layer [kg/s²]
            and higher orders.
        phonon_damping (float): damping constant of phonon propagation [kg/s].
        opt_pen_depth (float): optical penetration depth of the layer [m].
        opt_ref_index (ndarray[float]): optical refractive index - real
           and imagenary part :math:`n + i\kappa`.
        opt_ref_index_per_strain (ndarray[float]): optical refractive
           index change per strain - real and imagenary part
           :math:`\frac{d n}{d \eta} + i\frac{d \kappa}{d \eta}`.
        therm_cond (list[@lambda]): list of HANDLES T-dependent thermal
           conductivity [W/(m K)].
        lin_therm_exp (list[@lambda]): list of T-dependent linear thermal
           expansion coefficient (relative).
        int_lin_therm_exp (list[@lambda]): list of T-dependent integrated
           linear thermal expansion coefficient.
        heat_capacity (list[@lambda]): list of T-dependent heat capacity
           function [J/(kg K)].
        int_heat_capacity (list[@lambda]): list of T-dependent integrated heat
           capacity function.
        sub_system_coupling (list[@lambda]): list of of coupling functions of
           different subsystems [W/m³].
        num_sub_systems (int): number of subsystems for heat and phonons
           (electrons, lattice, spins, ...).
        atoms (list[atom, @lambda]): list of atoms and function handle
           for strain dependent displacement.
        num_atoms (int): number of atoms in unit cell.
        eff_spin (float): effective spin.
        curie_temp (float): Curie temperature [K].
        mf_exch_coupling (float): mean field exchange coupling constant [m²kg/s²].
        lamda (float): intrinsic coupling to bath parameter.
        mag_moment (float): atomic magnetic moment [mu_Bohr].
        aniso_exponent(ndarray[float]): exponent of T-dependence uniaxial
            anisotropy.
        anisotropy (float): anisotropy at T=0 K [J/m³] as x,y,z component vector.
        exch_stiffness (float): exchange stiffness at T=0 K [J/m].
        mag_saturation (float): saturation magnetization at 0 K [J/T/m³].
        magnetization (list[float]): magnetization amplitudes, phi, and
           gamma angle of each atom in the unit cell.

    """

    def __init__(self, id, name, c_axis, **kwargs):
        self.c_axis = c_axis
        self.thickness = c_axis
        self.a_axis = kwargs.get('a_axis', self.c_axis)
        self.b_axis = kwargs.get('b_axis', self.a_axis)
        self.mass = 0.0*u.kg
        self.mass_unit_area = 0.0*u.kg
        self.density = 0.0*u.kg/u.m**2

        super().__init__(id, name, **kwargs)

        self.area = self.a_axis * self.b_axis
        self.volume = self.area * self.c_axis
        self.atoms = []
        self.num_atoms = 0
        self.magnetization = []

    def __str__(self):
        """String representation of this class"""
        output = [['id', self.id],
                  ['name', self.name],
                  ['a-axis', '{:.4g~P}'.format(self.a_axis.to('nm'))],
                  ['b-axis', '{:.4g~P}'.format(self.b_axis.to('nm'))],
                  ['c-axis', '{:.4g~P}'.format(self.c_axis.to('nm'))],
                  ['area', '{:.4g~P}'.format(self.area.to('nm**2'))],
                  ['volume', '{:.4g~P}'.format(self.volume.to('nm**3'))],
                  ['mass', '{:.4g~P}'.format(self.mass.to('kg'))],
                  ['mass per unit area', '{:.4g~P}'.format(self.mass_unit_area)],
                  ]
        output += super().__str__()

        class_str = 'Unit Cell with the following properties\n\n'
        class_str += tabulate(output, headers=['parameter', 'value'], tablefmt='rst',
                              colalign=('right',), floatfmt=('.2f', '.2f'))
        class_str += '\n\n' + str(self.num_atoms) + ' Constituents:\n'

        atoms_str = []
        for i in range(self.num_atoms):
            atoms_str.append([self.atoms[i][0].name,
                              '{:0.2f}'.format(self.atoms[i][1](0)),
                              self.atoms[i][2],
                              '',
                              self.atoms[i][0].mag_amplitude,
                              self.atoms[i][0].mag_phi.to('deg').magnitude,
                              self.atoms[i][0].mag_gamma.to('deg').magnitude,
                              ])
        class_str += tabulate(atoms_str, headers=['atom', 'position', 'position function',
                                                  'magn.', 'amplitude', 'phi [°]', 'gamma [°]'],
                              tablefmt='rst')
        return class_str

    def visualize(self, block=True, **kwargs):
        """visualize

        Allows for 3D presentation of unit cell by allow for a & b
        coordinate of atoms.

        Todo:
            use the avogadro project as plugin
        Todo:
            create unit cell from CIF file e.g. by xrayutilities plugin.
        Todo:
            visualize magnetization per atom

        Args:
            **kwargs (str): strain for manipulating unit cell visualization.

        """
        import matplotlib.pyplot as plt
        from matplotlib import cm

        strain = kwargs.get('strain', 0)

        colors = [cm.get_cmap('Set1')(x) for x in np.linspace(0, 1, self.num_atoms)]
        atom_ids = self.get_atom_ids()

        plt.figure()
        atoms_plotted = np.zeros_like(atom_ids)
        for j in range(self.num_atoms):
            if not atoms_plotted[atom_ids.index(self.atoms[j][0].id)]:
                label = self.atoms[j][0].id
                atoms_plotted[atom_ids.index(self.atoms[j][0].id)] = True
                plt.plot(1+j, self.atoms[j][1](strain), 'o',
                         markersize=10,
                         markeredgecolor=[0, 0, 0],
                         markerfacecolor=colors[atom_ids.index(self.atoms[j][0].id)],
                         label=label)
            else:
                label = '_nolegend_'
                plt.plot(1+j, self.atoms[j][1](strain), 'o',
                         markersize=10,
                         markeredgecolor=[0, 0, 0],
                         markerfacecolor=colors[atom_ids.index(self.atoms[j][0].id)],
                         label=label)

        plt.axis([0.1, self.num_atoms+0.9, -0.1, (1.1+strain)])
        plt.grid(True)

        plt.title('Strain: {:0.2f}%'.format(strain))
        plt.ylabel('relative Position')
        plt.xlabel('# Atoms')
        plt.legend()
        plt.show(block=block)

    def add_atom(self, atom, position):
        r"""add_atom

        Adds an AtomBase/AtomMixed at a relative position of the unit cell.

        Sort the list of atoms by the position at zero strain.

        Update the mass, density and spring constant of the unit cell
        automatically:

        .. math:: \kappa = m \cdot (v_s / c)^2

        Args:
            atom (Atom, AtomMixed): Atom or AtomMixed added to unit cell.
            position (float): relative position within unit cel [0 .. 1].

        """
        s = symbols('s')
        position_str = ''
        # test the input type of the position
        if isfunction(position):
            raise ValueError('Please use string representation of function!')
        elif isinstance(position, str):
            try:
                # backwards compatibility for direct lambda definition
                if ':' in position:
                    # strip lambda prefix
                    position = position.split(':')[1]
                position_str = position.strip()
                position = lambdify(s, position, modules='numpy')
            except Exception as e:
                print('String input for unit cell property ' + position + ' \
                    cannot be converted to function handle!')
                print(e)
        elif isinstance(position, (int, float)):
            position_str = str(position) + '*(1+s)'
            position = lambdify(s, position_str, modules='numpy')
        else:
            raise ValueError('Atom position input has to be a scalar, or string'
                             'which can be converted into a lambda function!')

        # add the atom at the end of the array
        self.atoms.append([atom, position, position_str])
        # sort list of atoms by position at zero strain
        self.atoms.sort(key=lambda x: x[1](0))
        # increase the number of atoms
        self.num_atoms = self.num_atoms + 1

        self.magnetization.append([atom.mag_amplitude, atom.mag_phi, atom.mag_gamma])

        self.mass = 0*u.kg
        for i in range(self.num_atoms):
            self.mass = self.mass + self.atoms[i][0].mass

        self.density = self.mass / self.volume
        # set mass per unit area
        self.mass_unit_area = self.mass * 1*u.angstrom**2 / self.area
        self.calc_spring_const()

    def add_multiple_atoms(self, atom, position, Nb):
        """add_multiple_atoms

        Adds multiple AtomBase/AtomMixed at a relative position of the
        unit cell.

        Args:
            atom (Atom, AtomMixed): Atom or AtomMixed added to unit cell.
            position (float): relative position within unit cel [0 .. 1].
            Nb (int): repetition of atoms.

        """
        for _ in range(int(Nb)):
            self.add_atom(atom, position)

    def get_atom_ids(self):
        """get_atom_ids

        Provides a list of atom ids within the unit cell.

        Returns:
            ids (list[str]): list of atom ids within unit cell

        """
        ids = []
        for i in range(self.num_atoms):
            if not self.atoms[i][0].id in ids:
                ids.append(self.atoms[i][0].id)

        return ids

    def get_atom_positions(self, *args):
        """get_atom_positions

        Calculates the relative positions of the atoms in the unit cell

        Returns:
            res (ndarray[float]): relative postion of the atoms within the unit
            cell.

        """
        if args:
            strain = float(np.asarray(args[0]).item())
        else:
            strain = 0.

        res = np.zeros([self.num_atoms])
        for i, atom in enumerate(self.atoms):
            res[i] = atom[1](strain)

        return res

    @property
    def a_axis(self):
        return Q_(self._a_axis, u.meter).to('nm')

    @a_axis.setter
    def a_axis(self, a_axis):
        self._a_axis = a_axis.to_base_units().magnitude

    @property
    def b_axis(self):
        return Q_(self._b_axis, u.meter).to('nm')

    @b_axis.setter
    def b_axis(self, b_axis):
        self._b_axis = b_axis.to_base_units().magnitude

    @property
    def c_axis(self):
        return Q_(self._c_axis, u.meter).to('nm')

    @c_axis.setter
    def c_axis(self, c_axis):
        self._c_axis = c_axis.to_base_units().magnitude



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# The MIT License (MIT)
# Copyright (c) 2020 Daniel Schick
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
# DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
# OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE
# OR OTHER DEALINGS IN THE SOFTWARE.

__all__ = ['make_hash_md5', 'make_hashable', 'm_power_x',
           'm_times_n', 'finderb', 'multi_gauss', 'convert_cartesian_to_polar',
           'convert_polar_to_cartesian']

__docformat__ = 'restructuredtext'

import hashlib
import numpy as np


def make_hash_md5(obj):
    """make_hash_md5


    Args:
        obj (any): anything that can be hashed.

    Returns:
        hash (str): hash from object.

    """
    hasher = hashlib.md5()
    hasher.update(repr(make_hashable(obj)).encode())
    return hasher.hexdigest()


def make_hashable(obj):
    """make_hashable

    Recursive calls to elements of tuples, lists, dicts, set, and frozensets.

    Args:
        obj (any): anything that can be hashed..

    Returns:
        obj (tuple): hashable object.

    """
    if isinstance(obj, (tuple, list)):
        return tuple((make_hashable(e) for e in obj))

    if isinstance(obj, dict):
        return tuple(sorted((k, make_hashable(v)) for k, v in obj.items()))

    if isinstance(obj, (set, frozenset)):
        return tuple(sorted(make_hashable(e) for e in obj))

    return obj


def m_power_x(m, x):
    """m_power_x

    Apply ``numpy.linalg.matrix_power`` to last 2 dimensions of 4-dimensional
    input matrix.

    Args:
        m (ndarray[float, complex]): 4-dimensional input matrix.
        x (float): exponent.

    Returns:
        m (ndarray[float, complex]): resulting matrix.

    """
    if x > 1:
        for i in range(np.size(m, 0)):
            for j in range(np.size(m, 1)):
                m[i, j, :, :] = np.linalg.matrix_power(m[i, j, :, :], x)
    return m


def m_times_n(m, n):
    """m_times_n

    Matrix multiplication of last 2 dimensions for two 4-dimensional input
    matrices.

    Args:
        m (ndarray[float, complex]): 4-dimensional input matrix.
        n (ndarray[float, complex]): 4-dimensional input matrix.

    Returns:
        res (ndarray[float, complex]): 4-dimensional multiplication result.

    """
    return np.einsum('lmij,lmjk->lmik', m, n)


def finderb(key, array):
    """finderb

    Binary search algorithm for sorted array. Searches for the first index
    ``i`` of array where ``key`` >= ``array[i]``. ``key`` can be a scalar or
    a np.ndarray of keys. ``array`` must be a sorted np.ndarray.

    Author: André Bojahr.
    Licence: BSD.

    Args:
        key (float, ndarray[float]): single or multiple sorted keys.
        array (ndarray[float]): sorted array.

    Returns:
        i (ndarray[float]): position indices for each key in the array.

    """
    key = np.array(key, ndmin=1)
    n = len(key)
    i = np.zeros([n], dtype=int)

    for m in range(n):
        i[m] = finderb_nest(key[m], array)
    return i


def finderb_nest(key, array):
    """finderb_nest

    Nested sub-function of :func:`.finderb` for one single key.

    Author: André Bojahr.
    Licence: BSD.

    Args:
        key (float): single key.
        array (ndarray[float]): sorted array.

    Returns:
        a (float): position index of key in the array.

    """
    a = 0  # start of intervall
    b = len(array)  # end of intervall

    # if the key is smaller than the first element of the
    # vector we return 1
    if key < array[0]:
        return 0

    while (b-a) > 1:  # loop until the intervall is larger than 1
        c = int(np.floor((a+b)/2))  # center of intervall
        if key < array[c]:
            # the key is in the left half-intervall
            b = c
        else:
            # the key is in the right half-intervall
            a = c

    return a


def multi_gauss(x, s=[1], x0=[0], A=[1]):
    """multi_gauss

    Multiple gauss functions with width ``s`` given as FWHM and area normalized
    to input ``A`` and maximum of gauss at ``x0``.

    Args:
        x (ndarray[float]): argument of multi_gauss.
        s (ndarray[float], optional): FWHM of Gaussians. Defaults to 1.
        x0 (ndarray[float], optional): centers of Gaussians. Defaults to 0.
        A (ndarray[float], optional): amplitudes of Gaussians. Defaults to 1.

    Returns:
        y (ndarray[float]): multiple Gaussians.

    """
    s = np.asarray(s)/(2*np.sqrt(2*np.log(2)))
    a = np.asarray(A)/np.sqrt(2*np.pi*s**2)  # normalize area to 1
    x0 = np.asarray(x0)

    y = np.zeros_like(x)
    for i in range(len(s)):
        y = y + a[i] * np.exp(-((x-x0[i])**2)/(2*s[i]**2))
    return y


def convert_polar_to_cartesian(polar):
    r"""convert_polar_to_cartesian

    Convert a vector or field from polar coordinates
    :math:`(r, \phi, \gamma)` to cartesian coordinates :math:`(x, y, z)`:

    .. math::

        F_x & = r \sin(\phi)\cos(\gamma) \\
        F_y & = r \sin(\phi)\sin(\gamma) \\
        F_z & = r \cos(\phi)

    where :math:`r`, :math:`\phi`, :math:`\gamma` are the radius
    (amplitude), azimuthal, and polar angles of vector field
    :math:`\mathbf{F}`, respectively.

    Args:
        polar (ndarray[float]): vector of field to convert.

    Returns:
        cartesian (ndarray[float]): converted vector or field.

    """
    cartesian = np.zeros_like(polar)

    amplitudes = polar[..., 0]
    phis = polar[..., 1]
    gammas = polar[..., 2]
    cartesian[..., 0] = amplitudes*np.sin(phis)*np.cos(gammas)
    cartesian[..., 1] = amplitudes*np.sin(phis)*np.sin(gammas)
    cartesian[..., 2] = amplitudes*np.cos(phis)

    return cartesian


def convert_cartesian_to_polar(cartesian):
    r"""convert_cartesian_to_polar

    Convert a vector or field from cartesian coordinates :math:`(x, y, z)`
    to polar coordinates :math:`(r, \phi, \gamma)`:

    .. math::

        F_r & = \sqrt{F_x^2 + F_y^2+F_z^2}\\
        F_{\phi} & = \begin{cases}\\
        \arctan\left(\frac{F_y}{F_x} \right) & \mathrm{for}\ F_x > 0 \\
        \pi + \arctan\left(\frac{F_y}{F_x}\right)
        & \mathrm{for}\ F_x < 0 \ \mathrm{and}\ F_y \geq 0 \\
        \arctan\left(\frac{F_y}{F_x}\right) - \pi
        & \mathrm{for}\ F_x < 0 \ \mathrm{and}\ F_y < 0 \\
        0 & \mathrm{for}\ F_x = F_y = 0
        \end{cases} \\
        F_{\gamma} & = \arccos\left(\frac{F_z}{F_r} \right)

    where :math:`F_r`, :math:`F_{\phi}`, :math:`F_{\gamma}` are the radial
    (amplitude), azimuthal, and polar component of vector field
    :math:`\mathbf{F}`, respectively.

    Args:
        cartesian (ndarray[float]): vector of field to convert.

    Returns:
        polar (ndarray[float]): converted vector or field.

    """
    polar = np.zeros_like(cartesian)
    xs = cartesian[..., 0]
    ys = cartesian[..., 1]
    zs = cartesian[..., 2]
    amplitudes = np.sqrt(xs**2 + ys**2 + zs**2)
    mask = amplitudes != 0.  # mask for non-zero amplitudes
    polar[..., 0] = amplitudes
    polar[mask, 1] = np.arccos(np.divide(zs[mask], amplitudes[mask]))
    polar[..., 2] = np.arctan2(ys, xs)

    return polar



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from udkm1Dsim import Atom, AtomMixed
from udkm1Dsim import Layer, AmorphousLayer, UnitCell
from udkm1Dsim import Structure
from udkm1Dsim import u
import pytest


# fixtures

# atoms

@pytest.fixture(scope='module')
def atom_iron():
    atom_iron = Atom('Fe')
    atom_iron.mag_amplitude = 0.5
    atom_iron.mag_phi = 0*u.deg
    atom_iron.mag_gamma = 180*u.deg
    return atom_iron


@pytest.fixture(scope='module')
def atom_oxygen():
    atom_oxygen = Atom('O')
    atom_oxygen.ionicity = 1
    atom_oxygen.mag_amplitude = 0.5
    atom_oxygen.mag_phi = 0*u.deg
    atom_oxygen.mag_gamma = 180*u.deg
    return atom_oxygen


@pytest.fixture(scope='module')
def atom_strontium():
    atom_strontium = Atom('Sr')
    atom_strontium.mag_amplitude = 0.5
    atom_strontium.mag_phi = 0*u.deg
    atom_strontium.mag_gamma = 180*u.deg
    return atom_strontium


@pytest.fixture(scope='module')
def atom_titanium():
    atom_titanium = Atom('Ti')
    atom_titanium.mag_amplitude = 0.5
    atom_titanium.mag_phi = 0*u.deg
    atom_titanium.mag_gamma = 180*u.deg
    return atom_titanium


@pytest.fixture(scope='module')
def atom_dysprosium():
    atom_dysprosium = Atom('Dy')
    atom_dysprosium.ionicity = 1
    atom_dysprosium.mag_amplitude = 0.
    atom_dysprosium.mag_phi = 0.*u.deg
    atom_dysprosium.mag_gamma = 0.*u.deg
    return atom_dysprosium


@pytest.fixture(scope='module')
def atom_mixed(atom_dysprosium, atom_iron):
    atom_mixed = AtomMixed('DyFe')
    atom_mixed.add_atom(atom_dysprosium, 0.4)
    atom_mixed.add_atom(atom_iron, 0.6)
    atom_mixed.mag_amplitude = 0.
    atom_mixed.mag_phi = 0.*u.deg
    atom_mixed.mag_gamma = 0.*u.deg
    return atom_mixed


# layers


@pytest.fixture(scope='module')
def properties():
    props = {'roughness': 0.5*u.angstrom,
             'deb_wal_fac': 1*u.angstrom**2,
             'sound_vel': 6*u.nm/u.ps,
             'phonon_damping': 1*u.kg/u.s,
             'opt_pen_depth': 10.0*u.nm,
             'opt_ref_index': 5-3j,
             'opt_ref_index_per_strain': 1-1j,
             'therm_cond': 1*u.W/(u.m*u.K),
             'lin_therm_exp': 1e-5,
             'heat_capacity': 10*(u.J/u.kg/u.K),
             'sub_system_coupling': [0],
             'eff_spin': 1,
             'curie_temp': 100*u.K,
             'lamda': 1,
             'mag_moment': 1*u.bohr_magneton,
             'aniso_exponent': 1,
             'anisotropy': [1, 2, 3]*u.J/u.m**3,
             'exch_stiffness': 1*u.J/u.m,
             'mag_saturation': 1*u.J/u.T/u.m**3,
             }
    return props


@pytest.fixture(scope='module')
def layer(properties):
    return Layer(id='layer', name='base layer', **properties)


@pytest.fixture(scope='module')
def amorphous_layer(atom_iron, properties):
    return AmorphousLayer(id='amorphous_layer', name='amorphous layer', thickness=1*u.nm,
                          density=5000*u.kg/u.m**3, atom=atom_iron, **properties)


@pytest.fixture(scope='module')
def unit_cell(atom_strontium, atom_oxygen, atom_titanium, properties):
    uc = UnitCell(id='unit_cell', name='unit_cell', c_axis=5.0*u.angstrom, **properties)
    uc.add_atom(atom_strontium, 0.0)
    uc.add_atom(atom_oxygen, 0.5)
    uc.add_atom(atom_titanium, 1.0)
    return uc


@pytest.fixture(scope='module')
def amorphous_layer_iron(atom_iron, properties):
    return AmorphousLayer(id='amorphous_layer_Fe', name='amorphous layer iron', thickness=1*u.nm,
                          density=5000*u.kg/u.m**3, atom=atom_iron, **properties)


@pytest.fixture(scope='module')
def amorphous_layer_oxygen(atom_oxygen, properties):
    return AmorphousLayer(id='amorphous_layer_O', name='amorphous layer oxygen', thickness=1*u.nm,
                          density=5000*u.kg/u.m**3, atom=atom_oxygen, **properties)


@pytest.fixture(scope='module')
def unit_cell_iron(atom_iron, properties):
    uc = UnitCell(id='unit_cell_Fe', name='unit cell iron', c_axis=5.0*u.angstrom, **properties)
    uc.add_atom(atom_iron, 0.0)
    return uc


@pytest.fixture(scope='module')
def unit_cell_oxygen(atom_oxygen, properties):
    uc = UnitCell(id='unit_cell_O', name='unit cell oxygen', c_axis=5.0*u.angstrom, **properties)
    uc.add_atom(atom_oxygen, 0.0)
    return uc


# structure


@pytest.fixture(scope='module')
def structure(amorphous_layer_iron, amorphous_layer_oxygen, unit_cell_iron, unit_cell_oxygen):
    S = Structure('structure mixed')
    S.add_sub_structure(amorphous_layer_iron, 10)
    S.add_sub_structure(amorphous_layer_oxygen, 10)
    S.add_sub_structure(unit_cell_iron, 20)

    DL = Structure('double layer')
    DL.add_sub_structure(unit_cell_iron, 5)
    DL.add_sub_structure(unit_cell_oxygen, 7)

    S.add_sub_structure(DL, 20)

    substrate = Structure('substrate')
    substrate.add_sub_structure(amorphous_layer_iron, 100)

    S.add_substrate(substrate)

    return S


@pytest.fixture(scope='module')
def structure_amorph(amorphous_layer_iron, amorphous_layer_oxygen):
    S = Structure('structure amorph')
    S.add_sub_structure(amorphous_layer_iron, 10)
    S.add_sub_structure(amorphous_layer_oxygen, 10)

    substrate = Structure('substrate')
    substrate.add_sub_structure(amorphous_layer_iron, 100)

    S.add_substrate(substrate)

    return S


@pytest.fixture(scope='module')
def structure_crystalline(unit_cell_iron, unit_cell_oxygen):
    S = Structure('structure crystalline')
    S.add_sub_structure(unit_cell_iron, 10)
    S.add_sub_structure(unit_cell_oxygen, 10)

    substrate = Structure('substrate')
    substrate.add_sub_structure(unit_cell_iron, 100)

    S.add_substrate(substrate)

    return S



from setuptools import setup, find_packages

setup(
    name='udkm1Dsim',
    version='2.3.0',
    packages=find_packages(),
    package_data={
        'udkm1Dsim': ['parameters/atomic_form_factors/chantler/*.cf',
                      'parameters/atomic_form_factors/chantler/*.md',
                      'parameters/atomic_form_factors/henke/*.nff',
                      'parameters/atomic_form_factors/henke/*.md',
                      'parameters/atomic_form_factors/cromermann.txt',
                      'parameters/magnetic_form_factors/*.mf',
                      'parameters/elements.dat',
                      'matlab/*.m',
                      ],
    },
    url='https://github.com/dschick/udkm1Dsim',
    install_requires=['tqdm>=4.43.0',
                      'numpy',
                      'pint>=0.23',
                      'scipy>=1.4.1',
                      'sympy>=1.5.1',
                      'tabulate',
                      'matplotlib>=2.0.0'],
    extras_require={
        'parallel':  ['dask[distributed]>=2.6.0'],
        'testing': ['flake8', 'pytest'],
    },
    license='MIT',
    author='Daniel Schick',
    author_email='schick.daniel@gmail.com',
    description='A Python Simulation Toolkit for 1D Ultrafast Dynamics '
                + 'in Condensed Matter',
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    python_requires='>=3.10',
    keywords='ultrafast dynamics condensed matter 1D '
            + 'NTM 3TM 2TM TTM N-temperature model '
            + 'coherent acoustic phonons sound strain waves '
            + 'resonant magnetic scattering diffraction spectroscopy '
            + 'x-ray magnetic circular dichroism '
            + 'Landau Lifschitz Bloch ',
)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from udkm1Dsim import Magnetization, LLB
from udkm1Dsim import u
import numpy as np
import pytest


# fixtures


@pytest.fixture(scope='module')
def magnetization(structure, tmp_path_factory):
    return Magnetization(structure, force_recalc=True, cache_dir=tmp_path_factory.mktemp('cache'),
                         save_data=True, disp_messages=True, progress_bar=True,
                         )


@pytest.fixture(scope='module')
def llb(structure_amorph, tmp_path_factory):
    return LLB(structure_amorph, force_recalc=True, cache_dir=tmp_path_factory.mktemp('cache'),
               save_data=True, disp_messages=True, progress_bar=True,
               )


@pytest.fixture(scope='module')
def delays():
    return np.r_[-1:10:1]*u.ps


@pytest.fixture(scope='module')
def distances(structure_amorph):
    dists, _, _ = structure_amorph.get_distances_of_layers()
    return dists


@pytest.fixture(scope='module')
def temp_map(delays, distances):
    return 50*np.ones([len(delays), len(distances), 1])


# tests


# Magnetization


def test_magnetization_str(magnetization):
    magnetization.__str__()


def test_magnetization_get_hash(magnetization, delays, temp_map):
    magnetization.get_hash(delays=delays, temp_map=temp_map)


def test_magnetization_check_initial_magnetization(magnetization):
    # must fail as UnitCells are not yet supported, see #129
    with pytest.raises(TypeError):
        magnetization.check_initial_magnetization([])


# LLB


def test_llb_str(llb):
    llb.__str__()


def test_llb_check_initial_magnetization(llb, distances):
    llb.check_initial_magnetization(np.array([0, 0, 0]), distances=distances)
    llb.check_initial_magnetization(np.zeros([len(distances), 3]), distances=distances)
    with pytest.raises(ValueError):
        llb.check_initial_magnetization(np.zeros([3, len(distances)]),
                                        distances=distances)


def test_llb_get_magnetization(llb, temp_map, delays):
    llb.get_magnetization_map(delays, temp_map=temp_map)
    llb.force_recalc = False
    llb.get_magnetization_map(delays, temp_map=temp_map)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# The MIT License (MIT)
# Copyright (c) 2020 Daniel Schick
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
# DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
# OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE
# OR OTHER DEALINGS IN THE SOFTWARE.

__all__ = ['Phonon', 'PhononNum', 'PhononAna']

__docformat__ = 'restructuredtext'

from .simulation import Simulation
from ..helpers import make_hash_md5, finderb
import numpy as np
from os import path
from time import time
from scipy.integrate import solve_ivp
from tqdm.auto import tqdm, trange


class Phonon(Simulation):
    """Phonon

    Base class for phonon simulations in a linear chain of masses and springs.

    Args:
        S (Structure): sample to do simulations with.
        force_recalc (boolean): force recalculation of results.

    Keyword Args:
        save_data (boolean): true to save simulation results.
        cache_dir (str): path to cached data.
        disp_messages (boolean): true to display messages from within the
            simulations.
        progress_bar (boolean): enable tqdm progress bar.
        only_heat (boolean): true when including only thermal expansion without
            coherent phonon dynamics.

    Attributes:
        S (Structure): sample structure to calculate simulations on.
        force_recalc (boolean): force recalculation of results.
        save_data (boolean): true to save simulation results.
        cache_dir (str): path to cached data.
        disp_messages (boolean): true to display messages from within the
            simulations.
        progress_bar (boolean): enable tqdm progress bar.
        only_heat (boolean): true when including only thermal expansion without
            coherent phonon dynamics.

    """

    def __init__(self, S, force_recalc, **kwargs):
        super().__init__(S, force_recalc, **kwargs)
        self.only_heat = kwargs.get('only_heat', False)

    def __str__(self, output=[]):
        """String representation of this class"""

        output = [['only heat', self.only_heat],
                  ] + output

        class_str = 'Phonon simulation properties:\n\n'
        class_str += super().__str__(output)

        return class_str

    def get_hash(self, delays, temp_map, delta_temp_map, **kwargs):
        """get_hash

        Calculates an unique hash given by the ``delays``, and ``temp_map``
        and ``delta_temp_map`` as well as the sample structure hash for
        relevant lattice parameters.

        Args:
            delays (ndarray[float]): delay grid for the simulation.
            temp_map (ndarray[float]): spatio-temporal temperature profile.
            delta_temp_map (ndarray[float]): spatio-temporal temperature
                difference profile.
            **kwargs (float, optional): optional parameters.

        Returns:
            hash (str): unique hash.

        """
        param = [delays, self.only_heat]

        if np.size(temp_map) > 1e6:
            temp_map = temp_map.flatten()[0:1000000]
            delta_temp_map = delta_temp_map.flatten()[0:1000000]
        param.append(temp_map)
        param.append(delta_temp_map)

        for value in kwargs.values():
            param.append(value)

        return self.S.get_hash(types='phonon') + '_' + make_hash_md5(param)

    def get_all_strains_per_unique_layer(self, strain_map):
        """get_all_strains_per_unique_layer

        Determines all values of the strain per unique layer.

        Args:
            strain_map (ndarray[float]): spatio-temporal strain profile.

        Returns:
            strains (list[ndarray[float]]: all strains per unique layer.

        """
        # get the position indices of all unique layers in the sample structure
        positions = self.S.get_all_positions_per_unique_layer()
        strains = []

        for value in positions.values():
            strains.append(np.sort(np.unique(strain_map[:, value].flatten())))

        return strains

    def get_reduced_strains_per_unique_layer(self, strain_map, N=100):
        """get_reduced_strains_per_unique_layer

        Calculates all strains per unique layer that are given by the
        input ``strain_map``, but with a reduced number. The reduction is done
        by equally spacing the strains between the ``min`` and ``max`` strain
        with a given number :math:`N`, which can be also an array of the
        :math:`len(N) = L`, where :math:`L` is the number of unique layers.

        Args:
            strain_map (ndarray[float]): spatio-temporal strain profile.
            N (int, optional): number of reduced strains. Defaults to 100.

        Returns:
            strains (list[ndarray[float]]: reduced strains per unique layer.

        """
        # initialize
        all_strains = self.get_all_strains_per_unique_layer(strain_map)
        L = len(all_strains)  # Nb. of unique layers
        strains = []

        # normalize N to a 1D integer array of length L
        if np.size(N) == 1:
            N = np.full(L, int(N))
        else:
            N = np.asarray(N).ravel()
            if N.size != L:
                raise ValueError('The dimension of N must be either 1 or the number '
                                 'of unique layers ({:d}) the structure!'.format(L))
            N = N.astype(int)

        for i, value in enumerate(all_strains):
            min_strain = np.min(value)
            max_strain = np.max(value)

            strains.append(np.sort(np.unique(
                np.r_[0, np.linspace(min_strain, max_strain, int(N[i]))])))

        return strains

    def check_temp_maps(self, temp_map, delta_temp_map, delays):
        """check_temp_maps

        Check temperature profiles for correct dimensions.

        Args:
            temp_map (ndarray[float]): spatio-temporal temperature profile.
            delta_temp_map (ndarray[float]): spatio-temporal temperature
                difference profile.
            delays (ndarray[float]): delay grid for the simulation.

        Returns:
            (tuple):
            - *temp_map (ndarray[float])* - checked spatio-temporal temperature
              profile.
            - *delta_temp_map (ndarray[float])* - checked spatio-temporal
              differential temperature profile.

        """
        M = len(delays)
        L = self.S.get_number_of_layers()
        K = self.S.num_sub_systems

        # check size of delta_temp_map
        if K == 1:
            if np.shape(delta_temp_map) == (1, L):
                temp = delta_temp_map
                delta_temp_map = np.zeros([M, L])
                delta_temp_map[0, :] = temp
            elif (np.size(delta_temp_map, 0) != M) or (np.size(delta_temp_map, 1) != L):
                raise ValueError('The given temperature difference map does not have the '
                                 'dimension M x L, where M is the number of delay steps '
                                 'and L the number of layers!')
        else:
            if np.shape(delta_temp_map) == (1, L, K):
                temp = delta_temp_map
                delta_temp_map = np.zeros([K, L, K])
                delta_temp_map[0, :, :] = temp
            elif ((np.size(delta_temp_map, 0) != M)
                  or (np.size(delta_temp_map, 1) != L)
                  or (np.size(delta_temp_map, 2) != K)):
                raise ValueError('The given temperature difference map does not have the '
                                 'dimension M x L x K, where M is the number of delay steps '
                                 'and L the number of layers and K is the number of subsystems!')

        if np.shape(temp_map) != np.shape(delta_temp_map):
            raise ValueError('The temperature map does not have the same size as the '
                             'temperature difference map!')

        return temp_map, delta_temp_map

    def calc_sticks_from_temp_map(self, temp_map, delta_temp_map):
        r"""calc_sticks_from_temp_map

        Calculates the sticks to insert into the layer springs which model the
        external force (thermal stress). The length of :math:`l_i` of the
        :math:`i`-th spacer stick is calculated from the temperature-dependent
        linear thermal expansion :math:`\alpha(T)` of the layer:

        .. math::

            \alpha(T) = \frac{1}{L} \frac{d L}{d T}

        which results after integration in

        .. math::

            l = \Delta L = L_1 \exp(A(T_2) - A(T_1)) - L_1

        where :math:`A(T)` is the integrated lin. therm. expansion coefficient
        in respect to the temperature :math:`T`. The indices 1 and 2 indicate
        the initial and final state.

        Args:
            temp_map (ndarray[float]): spatio-temporal temperature profile.
            delta_temp_map (ndarray[float]): spatio-temporal temperature
                difference profile.

        Returns:
            (tuple):
            - *sticks (ndarray[float])* - summed spacer sticks.
            - *sticks_sub_systems (ndarray[float])* - spacer sticks per
              sub-system.

        """
        M = np.size(temp_map, 0)  # nb of delay steps
        L = self.S.get_number_of_layers()
        K = self.S.num_sub_systems

        temp_map = np.reshape(temp_map, [M, L, K])
        delta_temp_map = np.reshape(delta_temp_map, [M, L, K])

        thicknesses = self.S.get_layer_property_vector('_thickness')
        int_lin_therm_exps = self.S.get_layer_property_vector('_int_lin_therm_exp')

        # evaluated initial integrated linear thermal expansion from T1 to T2
        int_alpha_T0 = np.zeros([L, K])
        # evaluated integrated linear thermal expansion from T1 to T2
        int_alpha_T = np.zeros([L, K])
        sticks = np.zeros([M, L])  # the sticks inserted in the unit cells
        sticks_sub_systems = np.zeros([M, L, K])  # the sticks for each thermodynamic subsystem

        # calculate initial integrated linear thermal expansion from T1 to T2
        # traverse subsystems
        for ii in range(L):
            for iii in range(K):
                int_alpha_T0[ii, iii] = int_lin_therm_exps[ii][iii](temp_map[0, ii, iii]
                                                                    - delta_temp_map[0, ii, iii])

        # calculate sticks for all subsystems for all delay steps
        # traverse time
        for i in range(M):
            if np.any(delta_temp_map[i, :]):  # there is a temperature change
                # Calculate new sticks from the integrated linear thermal
                # expansion from initial temperature to current temperature for
                # each subsystem
                # traverse subsystems
                for ii in range(L):
                    for iii in range(K):
                        int_alpha_T[ii, iii] = int_lin_therm_exps[ii][iii](temp_map[i, ii, iii])

                # calculate the length of the sticks of each subsystem and sum
                # them up
                sticks_sub_systems[i, :, :] = np.tile(thicknesses, (K, 1)).T \
                    * np.exp(int_alpha_T-int_alpha_T0) - np.tile(thicknesses, (K, 1)).T
                sticks[i, :] = np.sum(sticks_sub_systems[i, :, :], 1)
            else:  # no temperature change, so keep the current sticks
                if i > 0:
                    sticks_sub_systems[i, :, :] = sticks_sub_systems[i-1, :, :]
                    sticks[i, :] = sticks[i-1, :]
        return sticks, sticks_sub_systems


class PhononNum(Phonon):
    """PhononNum

    Numerical model to simulate coherent acoustic phonons.

    Args:
        S (Structure): sample to do simulations with.
        force_recalc (boolean): force recalculation of results.

    Keyword Args:
        save_data (boolean): true to save simulation results.
        cache_dir (str): path to cached data.
        disp_messages (boolean): true to display messages from within the
            simulations.
        progress_bar (boolean): enable tqdm progress bar.
        only_heat (boolean): true when including only thermal expansion without
            coherent phonon dynamics.

    Attributes:
        S (Structure): sample structure to calculate simulations on.
        force_recalc (boolean): force recalculation of results.
        save_data (boolean): true to save simulation results.
        cache_dir (str): path to cached data.
        disp_messages (boolean): true to display messages from within the
            simulations.
        progress_bar (boolean): enable tqdm progress bar.
        only_heat (boolean): true when including only thermal expansion without
            coherent phonon dynamics.
        ode_options (dict): options for scipy solve_ivp ode solver.

    References:

        .. [7] A. Bojahr, M. Herzog, D. Schick, I. Vrejoiu, & M. Bargheer,
           *Calibrated real-time detection of nonlinearly propagating
           strain waves*, `Phys. Rev. B, 86(14), 144306 (2012).
           <http://www.doi.org/10.1103/PhysRevB.86.144306>`_

    """

    def __init__(self, S, force_recalc, **kwargs):
        super().__init__(S, force_recalc, **kwargs)
        self.ode_options = {
            'method': 'RK23',
            'first_step': None,
            'max_step': np.inf,
            'rtol': 1e-3,
            'atol': 1e-6,
            }

    def __str__(self, output=[]):
        """String representation of this class"""

        class_str = 'Numerical Phonon simulation properties:\n\n'
        class_str += super().__str__()

        return class_str

    def get_strain_map(self, delays, temp_map, delta_temp_map):
        """get_strain_map

        Returns a strain profile for the sample structure for given temperature
        profile. The result can be saved using an unique hash of the sample
        and the simulation parameters in order to reuse it.

        Args:
            delays (ndarray[Quantity]): delays range of simulation [s].
            temp_map (ndarray[float]): spatio-temporal temperature map.
            delta_temp_map (ndarray[float]): spatio-temporal differential
            temperature map.

        Returns:
            strain_map (ndarray[float]): spatio-temporal strain profile.

        """
        filename = 'strain_map_num_' \
                   + self.get_hash(delays, temp_map, delta_temp_map) \
                   + '.npz'
        full_filename = path.abspath(path.join(self.cache_dir, filename))
        if path.exists(full_filename) and not self.force_recalc:
            # found something so load it
            tmp = np.load(full_filename)
            strain_map = tmp['strain_map']
            self.disp_message('_strain_map_ loaded from file:\n\t' + filename)
        else:
            # file does not exist so calculate and save
            strain_map, _, _ = \
                self.calc_strain_map(delays, temp_map, delta_temp_map)
            self.save(full_filename, {'strain_map': strain_map}, '_strain_map_num_')
        return strain_map

    def calc_strain_map(self, delays, temp_map, delta_temp_map):
        r"""calc_strain_map

        Calculates the ``strain_map`` of the sample structure for a given
        ``temp_map`` and ``delta_temp_map`` and ``delay`` array. Further
        details are given in Ref. [7]_. The coupled differential equations are
        solved for each oscillator in a linear chain of masses and springs:

        .. math::

            m_i\ddot{x}_i = & -k_i(x_i-x_{i-1})-k_{i+1}(x_i-x_{i+1}) \\
            & + m_i\gamma_i(\dot{x}_i-\dot{x}_{i-1}) + F_i^{heat}(t)

        where :math:`x_i(t) = z_{i}(t) - z_i^0` is the shift of each layer.
        :math:`m_i` is the mass and :math:`k_i = m_i\, v_i^2/c_i^2` is the
        spring constant of each layer. Furthermore an empirical damping term
        :math:`F_i^{damp} = \gamma_i(\dot{x}_i-\dot{x}_{i-1})` is introduced
        and the external force (thermal stress) :math:`F_i^{heat}(t)`.
        The thermal stresses are modelled as spacer sticks which are
        calculated from the linear thermal expansion coefficients. The
        equation of motion can be reformulated as:

        .. math::

            m_i\ddot{x}_i = F_i^{spring} + F_i^{damp} + F_i^{heat}(t)

        The numerical solution also allows for non-harmonic inter-atomic
        potentials of up to the order :math:`M`. Accordingly
        :math:`k_i = (k_i^1 \ldots k_i^{M-1})` can be an array accounting for
        higher orders of the potential which is in the harmonic case purely
        quadratic (:math:`k_i = k_i^1`). The resulting force from the
        displacement of the springs

        .. math::

            F_i^{spring} = -k_i(x_i-x_{i-1})-k_{i+1}(x_i-x_{i+1})

        includes:

        .. math::

            k_i(x_i-x_{i-1}) = \sum_{j=1}^{M-1} k_i^j (x_i-x_{i-1})^j

        Args:
            delays (ndarray[Quantity]): delays range of simulation [s].
            temp_map (ndarray[float]): spatio-temporal temperature map.
            delta_temp_map (ndarray[float]): spatio-temporal differential
                temperature map.

        Returns:
            (tuple):
            - *strain_map (ndarray[float])* - spatio-temporal strain profile.
            - *sticks_sub_systems (ndarray[float])* - spacer sticks per
              sub-system.
            - *velocities (ndarray[float])* - spatio-temporal velocity profile.

        """
        t1 = time()

        # initialize
        L = self.S.get_number_of_layers()
        thicknesses = self.S.get_layer_property_vector('_thickness')
        x0 = np.zeros([2*L])  # initial condition for the shift of the layers

        try:
            delays = delays.to('s').magnitude
        except AttributeError:
            pass

        # check temp_maps
        [temp_map, delta_temp_map] = self.check_temp_maps(temp_map, delta_temp_map, delays)

        # calculate the sticks due to heat expansion first for all delay steps
        self.disp_message('Calculating linear thermal expansion ...')
        sticks, sticks_sub_systems = self.calc_sticks_from_temp_map(temp_map, delta_temp_map)

        if self.only_heat:
            # no coherent dynamics so calculate the strain directly
            strain_map = sticks/np.tile(thicknesses, [np.size(sticks, 0), 1])
            velocities = np.zeros_like(strain_map)  # this is quasi-static
        else:
            # include coherent dynamics
            self.disp_message('Calculating coherent dynamics with ODE solver ...')

            L = self.S.get_number_of_layers()
            masses = self.S.get_layer_property_vector('_mass_unit_area')
            thicknesses = self.S.get_layer_property_vector('_thickness')
            spring_consts = self.S.get_layer_property_vector('spring_const')
            damping = self.S.get_layer_property_vector('_phonon_damping')
            force_from_heat = PhononNum.calc_force_from_heat(sticks, spring_consts)

            # apply scipy's ode-solver together
            if self.progress_bar:  # with tqdm progressbar
                pbar = tqdm()
                pbar.set_description('Delay = {:.3f} ps'.format(delays[0]*1e12))
                state = [delays[0], abs(delays[-1]-delays[0])/100]
            else:  # without progressbar
                pbar = None
                state = None

            sol = solve_ivp(
                PhononNum.ode_func,
                [delays[0], delays[-1]],
                x0,
                args=(delays, force_from_heat, damping, spring_consts, masses, L,
                      pbar, state),
                t_eval=delays,
                **self.ode_options)

            if pbar is not None:  # close tqdm progressbar if used
                pbar.close()

            # calculate the strain_map as the second spacial derivative
            # of the layer shift x(t). The result of the ode solver
            # contains x(t) = X(:,1:N) and v(t) = X(:,N+1:end) the
            # positions and velocities of the layers, respectively.
            temp = np.diff(sol.y[0:L, :].T, 1, 1)
            strain_map = np.zeros([temp.shape[0], temp.shape[1]+1])
            strain_map[:, :-1] = temp
            strain_map = strain_map/np.tile(thicknesses[:], [np.size(temp, 0), 1])
            velocities = sol.y[L:, :].T
        self.disp_message('Elapsed time for _strain_map_:'
                          ' {:f} s'.format(time()-t1))
        return strain_map, sticks_sub_systems, velocities

    @staticmethod
    def ode_func(t, X, delays, force_from_heat, damping, spring_consts, masses, L,
                 pbar=None, state=None):
        r"""ode_func

        Provides the according ode function for the ode solver which has to be
        solved. The ode function has the input :math:`t` and :math:`X(t)` and
        calculates the temporal derivative :math:`\dot{X}(t)` where the vector

        .. math::

            X(t) = [x(t) \; \dot{x}(t)] \quad \mbox{and } \quad
            \dot{X}(t) = [\dot{x}(t) \; \ddot{x}(t)] .

        :math:`x(t)` is the actual shift of each layer.

        Args:
            t (ndarray[float]): internal time steps of the ode solver.
            X (ndarray[float]): internal variable of the ode solver.
            delays (ndarray[float]): delays range of simulation [s].
            force_from_heat (ndarray[float]): force due to thermal expansion.
            damping (ndarray[float]): phonon damping.
            spring_consts (ndarray[float]): spring constants of masses.
            masses (ndarray[float]): masses of layers.
            L (int): number of layers.
            pbar (tqdm): tqdm progressbar.
            state (list[float]): state variables for progress bar.

        Returns:
            X_prime (ndarray[float]): velocities and accelerations of masses.

        """
        if pbar is not None:
            # set everything for the tqdm progressbar
            last_t, dt = state
            n = (t - last_t)/dt
            if n >= 1:
                pbar.update(1)
                pbar.set_description('Delay = {:.3f} ps'.format(t*1e12))
                state[0] = t
            elif n < 0:
                state[0] = t

        # start with the actual ode function
        x = X[0:L]
        v = X[L:]

        # the output must be a column vector
        X_prime = np.zeros([2*L])

        # accelerations = derivative of velocities
        X_prime[L:] = (
            PhononNum.calc_force_from_damping(v, damping, masses)
            + PhononNum.calc_force_from_spring(
                np.r_[np.diff(x), 0],
                np.r_[0, np.diff(x)],
                spring_consts)
            + force_from_heat[:, finderb(t, delays)].squeeze()
            )/masses

        # velocities = derivative of positions
        X_prime[0:L] = v

        return X_prime

    @staticmethod
    def calc_force_from_spring(d_X1, d_X2, spring_consts):
        r"""calc_force_from_spring

        Calculates the force :math:`F_i^{spring}` acting on each mass due to
        the displacement between the left and right site of that mass.

        .. math::

            F_i^{spring} = -k_i(x_i-x_{i-1})-k_{i+1}(x_i-x_{i+1})

        We introduce-higher order inter-atomic potentials by

        .. math::

            k_i(x_i-x_{i-1}) = \sum_{j=1}^{M} k_i^j (x_i-x_{i-1})^j

        where :math:`M` is the order of the spring constants.

        Args:
            d_X1 (ndarray[float]): left displacements.
            d_X2 (ndarray[float]): right displacements.
            spring_consts (ndarray[float]): spring constants of masses.

        Returns:
            F (ndarray[float]): force from springs.

        """
        try:
            spring_order = np.size(spring_consts, 1)
        except IndexError:
            spring_order = 1

        spring_consts = np.reshape(spring_consts, [np.size(spring_consts, 0), spring_order])

        coeff1 = np.vstack((-spring_consts[0:-1, :], np.zeros([1, spring_order])))
        coeff2 = np.vstack((np.zeros([1, spring_order]), -spring_consts[0:-1, :]))

        temp1 = np.zeros([len(d_X1), spring_order])
        temp2 = np.zeros([len(d_X1), spring_order])

        for i in range(spring_order):
            temp1[:, i] = d_X1**(i+1)
            temp2[:, i] = d_X2**(i+1)

        F = np.sum(coeff2*temp2, 1) - np.sum(coeff1*temp1, 1)

        return F

    @staticmethod
    def calc_force_from_heat(sticks, spring_consts):
        """calc_force_from_heat

        Calculates the force acting on each mass due to the heat expansion,
        which is modeled by spacer sticks.

        Args:
            sticks (ndarray[float]): spacer sticks.
            spring_consts (ndarray[float]): spring constants of masses.

        Returns:
            F (ndarray[float]): force from thermal expansion.

        """
        M, L = np.shape(sticks)
        F = np.zeros([L, M])
        # traverse time
        for i in range(M):
            F[:, i] = -PhononNum.calc_force_from_spring(
                np.hstack((sticks[i, 0:L-1], 0)),
                np.hstack((0, sticks[i, 0:L-1])),
                spring_consts
                )

        return F

    @staticmethod
    def calc_force_from_damping(v, damping, masses):
        r"""calc_force_from_damping

        Calculates the force acting on each mass in a linear spring due to
        damping (:math:`\gamma_i`) according to the shift velocity difference
        :math:`v_{i}-v_{i-1}` with :math:`v_i(t) = \dot{x}_i(t)`:

        .. math::

            F_i^{damp} = \gamma_i(\dot{x}_i-\dot{x}_{i-1})

        Args:
            v (ndarray[float]): velocity of masses.
            damping (ndarray[float]): phonon damping.
            masses (ndarray[float]): masses.

        Returns:
            F (ndarray[float]): force from damping.

        """
        F = masses*damping*np.diff(v, 0)

        return F


class PhononAna(Phonon):
    """PhononAna

    Analytical model to simulate coherent acoustic phonons.

    Args:
        S (Structure): sample to do simulations with.
        force_recalc (boolean): force recalculation of results.

    Keyword Args:
        save_data (boolean): true to save simulation results.
        cache_dir (str): path to cached data.
        disp_messages (boolean): true to display messages from within the
            simulations.
        progress_bar (boolean): enable tqdm progress bar.
        only_heat (boolean): true when including only thermal expansion without
            coherent phonon dynamics.

    Attributes:
        S (Structure): sample structure to calculate simulations on.
        force_recalc (boolean): force recalculation of results.
        save_data (boolean): true to save simulation results.
        cache_dir (str): path to cached data.
        disp_messages (boolean): true to display messages from within the
            simulations.
        progress_bar (boolean): enable tqdm progress bar.
        only_heat (boolean): true when including only thermal expansion without
            coherent phonon dynamics.

    References:

        .. [8] M. Herzog, D. Schick, P. Gaal, R. Shayduk, C. von Korff Schmising
           & M. Bargheer, *Analysis of ultrafast X-ray diffraction data in a
           linear-chain model of the lattice dynamics*, `Applied Physics A,
           106(3), 489-499 (2011).
           <http://www.doi.org/doi:10.1007/s00339-011-6719-z>`_

    """

    def __init__(self, S, force_recalc, **kwargs):
        super().__init__(S, force_recalc, **kwargs)

    def __str__(self, output=[]):
        """String representation of this class"""

        class_str = 'Analytical Phonon simulation properties:\n\n'
        class_str += super().__str__()

        return class_str

    def get_strain_map(self, delays, temp_map, delta_temp_map):
        """get_strain_map

        Returns a strain profile for the sample structure for given temperature
        profile. The result can be saved using an unique hash of the sample
        and the simulation parameters in order to reuse it.

        Args:
            delays (ndarray[Quantity]): delays range of simulation [s].
            temp_map (ndarray[float]): spatio-temporal temperature map.
            delta_temp_map (ndarray[float]): spatio-temporal differential
            temperature map.

        Returns:
            (tuple):
            - *strain_map (ndarray[float])* - spatio-temporal strain profile.
            - *A (ndarray[float])* - coefficient vector A of general solution.
            - *B (ndarray[float])* - coefficient vector B of general solution.

        """
        filename = 'strain_map_ana_' \
                   + self.get_hash(delays, temp_map, delta_temp_map) \
                   + '.npz'
        full_filename = path.abspath(path.join(self.cache_dir, filename))
        if path.exists(full_filename) and not self.force_recalc:
            # found something so load it
            tmp = np.load(full_filename)
            strain_map = tmp['strain_map']
            A = tmp['A']
            B = tmp['B']
            self.disp_message('_strain_map_ loaded from file:\n\t' + filename)
        else:
            # file does not exist so calculate and save
            strain_map, A, B = \
                self.calc_strain_map(delays, temp_map, delta_temp_map)
            self.save(full_filename, {'strain_map': strain_map,
                                      'A': A,
                                      'B': B},
                      '_strain_map_ana_')
        return strain_map, A, B

    def calc_strain_map(self, delays, temp_map, delta_temp_map):
        r"""calc_strain_map

        Calculates the ``strain_map`` of the sample structure for a given
        ``temp_map`` and ``delta_temp_map`` and ``delay`` array. Further
        details are given in Ref. [8]_. Within the linear chain of :math:`N`
        masses (:math:`m_i`) at position :math:`z_i` coupled with spring
        constants :math:`k_i` one can formulate the differential equation
        of motion as follow:

        .. math::

            m_i\ddot{x}_i = -k_i(x_i-x_{i-1})-k_{i+1}(x_i-x_{i+1}) + F_i^{heat}(t)

        Since we only consider nearest-neighbor interaction one can write:

        .. math::

            \ddot{x}_i = \sum_{n=1}^N \kappa_{i,n} x_n = \Delta_i(t)

        Here :math:`x_i(t) = z_i(t)-z_i^0` is the shift of each layer,
        :math:`F_i^{heat}(t)` is the external force (thermal stress) of each
        layer and :math:`\kappa_{i,i} = -(k_i + k_{i+1})/m_i`, and
        :math:`\kappa_{i,i+1} = \kappa_{i+1,i} = k_{i+1}/m_i`.

        :math:`k_i = m_i\, v_i^2/c_i^2` is the spring constant and :math:`c_i`
        and :math:`v_i` are the thickness and longitudinal sound velocity of
        each layer respectively.
        One can rewrite the homogeneous differential equation in matrix
        form to obtain the general solution

        .. math::

            \frac{d^2}{dt^2} X = K \, X

        Here :math:`X = (x_1 \ldots x_N)` and :math:`K` is the
        tri-diagonal matrix of :math:`\kappa` which is real and symmetric.
        The differential equation can be solved with the ansatz:

        .. math::

         X(t) = \sum_j \Xi_j \, (A_j \cos(\omega_j \, t) + B_j \sin(\omega_j \, t))

        where :math:`\Xi_j = (\xi_1^j \ldots \xi_N^j)` are the eigenvectors of
        the matrix :math:`K`. Thus by solving the Eigenproblem for :math:`K` one
        gets the eigenvecotrs :math:`\Xi_j` and the eigenfrequencies
        :math:`\omega_j`. From the initial conditions

        .. math::

            X(0) = \sum_j \Xi_j \, A_j = \Xi \, A \qquad V(0) = \dot{X}(0)
                 = \sum_j \Xi_j \, \omega_j\, B_j = \Xi \, \omega \, B

        one can determine the real coefficient vectors :math:`A` and :math:`B` in
        order to calculate :math:`X(t)` and :math:`V(t)` using the ansatz:

        .. math::

            A = \Xi \setminus X(0) \qquad B = (\Xi \setminus V(0)) / \omega

        The external force is implemented as spacer sticks which are
        inserted into the springs and hence the layers have a new
        equilibrium positions :math:`z_i(\infty) = z_i^\infty`.
        Thus we can do a coordination transformation:

        .. math::

            z_i(t) = z_i^0 + x_i(t) = z_i^\infty + x_i^\infty(t)

        and

        .. math::

            x_i^\infty(t) = z_i^0 - z_i^\infty + x_i(t)

        with the initial condition :math:`x_i(0) = 0` it becomes

        .. math::

            x_i^\infty(0) = z_i^0 - z_i^\infty = \sum_{j = i+1}^N l_j

        :math:`x_i^\infty(0)` is the new initial condition after the excitation
        where :math:`l_i` is the length of the :math:`i`-th spacer stick.
        The spacer sticks are calculated from the temperature change and the
        linear thermal expansion coefficients.
        The actual strain :math:`\epsilon_i(t)` of each layer is calculates
        as follows:

        .. math::

            \epsilon_i(t) = [ \Delta x_i(t) + l_i) ] / c_i

        with :math:`\Delta x_i = x_i - x_{i-1}`. The stick :math:`l_i` have
        to be added here, because :math:`x_i` has been transformed into the new
        coordinate system :math:`x_i^\infty`.

        Args:
            delays (ndarray[Quantity]): delays range of simulation [s].
            temp_map (ndarray[float]): spatio-temporal temperature map.
            delta_temp_map (ndarray[float]): spatio-temporal differential
              temperature map.

        Returns:
            (tuple):
            - *strain_map (ndarray[float])* - spatio-temporal strain profile.
            - *A (ndarray[float])* - coefficient vector A of general solution.
            - *B (ndarray[float])* - coefficient vector B of general solution.

        """
        t1 = time()

        # initialize
        L = self.S.get_number_of_layers()
        M = len(delays)

        try:
            delays = delays.to('s').magnitude
        except AttributeError:
            pass

        delay0 = delays[0]  # initial delay
        thicknesses = self.S.get_layer_property_vector('_thickness')
        X = np.zeros([M, L])  # shifts of the layers
        V = np.zeros_like(X)  # velocities of the layers
        A = np.zeros_like(X)  # coefficient vector for eigenwert solution
        B = np.zeros_like(X)  # coefficient vector for eigenwert solution
        strain_map = np.zeros_like(X)  # the restulting strain map

        # check temp_maps
        [temp_map, delta_temp_map] = self.check_temp_maps(temp_map, delta_temp_map, delays)

        # calculate the sticks due to heat expansion first for all delay steps
        self.disp_message('Calculating linear thermal expansion ...')
        sticks, _ = self.calc_sticks_from_temp_map(temp_map, delta_temp_map)

        if self.only_heat:
            # no coherent dynamics so calculate the strain directly
            strain_map = sticks/np.tile(thicknesses, [np.size(sticks, 0), 1])
        else:
            # solve the eigenproblem for the structure to obtains the
            # eigenvectors X_i and eigenfreqeuencies omega for the L
            # coupled differential equations
            Xi, omega = self.solve_eigenproblem()
            # calculate the actual strain map with the solution of the
            # eigenproblem and the external force (sticks, thermal stress)
            self.disp_message('Calculating _strain_map_ ...')
            if self.progress_bar:
                iterator = trange(M, desc='Progress', leave=True)
            else:
                iterator = range(M)
            for i in iterator:
                dt = delays[i]-delay0  # this is the time step
                # calculate the current shift X and velocity V of all
                # layers using the ansatz
                X[i, :] = np.dot(Xi, (A[i, :].T*np.cos(omega*dt) + B[i, :].T*np.sin(omega*dt)))
                V[i, :] = np.dot(Xi, (omega*(-A[i, :].T*np.sin(omega*dt)
                                             + B[i, :].T*np.cos(omega*dt))))
                # remember the velocities and shifts as ic for the next
                # time step
                X0 = X[i, :].T
                V0 = V[i, :].T
                # the strain can only be calculated for L-1 layers, so
                # we neglect the last one
                if i > 0:
                    strain_map[i, 0:-1] = (np.diff(X[i, :])
                                           + sticks[i-1, 0:-1])/thicknesses[0:-1].T
                else:
                    # initial sticks are zero
                    strain_map[i, 0:-1] = np.diff(X[i, :])/thicknesses[0:-1].T
                # calculate everything for the next step
                if i < (M-1):  # check, if there is a next step
                    if np.any(delta_temp_map[i, :]):  # there is a temperature change
                        delay0 = delays[i]  # set new initial delay
                        # determining the shifts due to inserted sticks
                        # as new initial conditions
                        if i > 0:
                            temp = np.flipud(np.cumsum(np.flipud(sticks[i, :].T-sticks[i-1, :].T)))
                        else:
                            # initial sticks are zero
                            temp = np.flipud(np.cumsum(np.flipud(sticks[i, :].T)))
                        X0 = X0 + temp
                        # determining the coefficient vectors A and B of
                        # the general solution of X(t) using the initial
                        # conditions X0 and V0
                        A[i+1, :] = np.linalg.solve(Xi, X0)
                        B[i+1, :] = (np.linalg.solve(Xi, V0)/omega).T
                    else:
                        # no temperature change, so keep the current As,
                        # Bs, and sticks
                        A[i+1, :] = A[i, :]
                        B[i+1, :] = B[i, :]

        self.disp_message('Elapsed time for _strain_map_:'
                          ' {:f} s'.format(time()-t1))

        return strain_map, A, B

    def solve_eigenproblem(self):
        r"""solve_eigenproblem

        Creates the real and symmetric :math:`K` matrix (:math:`L \times L`) of
        spring constants :math:`k_i` and masses :math:`m_i` and calculates the
        eigenvectors :math:`\Xi_j` and eigenfrequencies :math:`\omega_j` for the
        matrix which are used to calculate the ``strain_map`` of the structure.
        If the result has been save to file, load it from there.

        Returns:
            (tuple):
            - *Xi (ndarray[float])* - eigenvectors.
            - *omega (ndarray[float])* - eigenfrequencies.

        """
        # create the file name to look for
        filename = 'eigenvalues_' \
                   + self.S.get_hash(types='phonon') \
                   + '.npz'
        full_filename = path.abspath(path.join(self.cache_dir, filename))
        if path.exists(full_filename) and not self.force_recalc:
            # found something so load it
            tmp = np.load(full_filename)
            Xi = tmp['Xi']
            omega = tmp['omega']
            self.disp_message('_eigen_values_ loaded from file:\n\t' + filename)
        else:
            # file does not exist so calculate and save
            t1 = time()
            self.disp_message('Calculating _eigen_values_ ...')
            # initialize
            L = self.S.get_number_of_layers()
            K = np.zeros([L, L])  # initializing three-diagonal springs-masses matrix.
            # initializing a vector for eigenfrequencies
            omega = np.zeros([L, 1], dtype=np.complex128)

            masses = self.S.get_layer_property_vector('_mass_unit_area')
            spring_consts = self.S.get_layer_property_vector('spring_const')
            spring_consts = np.hstack((0, spring_consts))  # set the first spring free

            for i in range(L):  # defining main diagonal
                K[i, i] = -(spring_consts[i] + spring_consts[i+1])/masses[i]

            # defining the two other diagonals - nearest neighbor interaction
            for i in range(1, L):
                K[i, i-1] = spring_consts[i]/masses[i]
                K[i-1, i] = spring_consts[i]/masses[i-1]

            # determining the eigenvectors and the eigenvalues
            lambd, Xi = np.linalg.eig(K)

            # calculate the eigenfrequencies from the eigenvalues
            omega = np.sqrt(-lambd)

            self.disp_message('Elapsed time for _eigen_values_:'
                              ' {:f} s'.format(time()-t1))
            # save the result to file
            self.save(full_filename, {'Xi': Xi, 'omega': omega}, '_eigen_values_')

        return Xi, omega

    def get_energy_per_eigenmode(self, A, B):
        r"""get_energy_per_eigenmode

        Returns the sorted energy per Eigenmode of the coherent phonons of
        the 1D sample.

        .. math::

            E_j = \frac{1}{2} (A^2_j + B^2_j)\, \omega_j^2\, m_j \, \| \Xi_j\|^2

        Frequencies are in [Hz] and energy per mode in [J].

        Args:
            A (ndarray[float]): coefficient vector A of general solution.
            B (ndarray[float]): coefficient vector B of general solution.

        Returns:
            (tuple):
            - *omega (ndarray[float])* - eigenfrequencies.
            - *E (ndarray[float])* - energy per eigenmode.

        """
        # initialize
        L = self.S.get_number_of_layers()
        M = A.shape[0]  # nb of delays
        E = np.zeros([M, L])
        masses = self.S.get_layer_property_vector('_mass_unit_area')

        # get the eigenVectors and eigenFrequencies
        Xi, omega = self.solve_eigenproblem()

        # sort the frequencies and remember the permutation of indices
        idx = np.argsort(omega)

        # traverse time
        for i in range(M):
            # calculate the energy for the jth mode
            E[i, :] = 0.5 * (A[i, :].T**2 + B[i, :].T**2) * omega**2 * masses * np.sum(Xi**2, 0).T

        return omega[idx], E[:, idx]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from udkm1Dsim import Simulation
import numpy as np
import pytest


# fixtures


@pytest.fixture(scope='module')
def simulation(structure):
    return Simulation(structure, force_recalc=True, cache_dir='./',
                      save_data=True, disp_messages=True, progress_bar=True)


# tests


def test_simulation_str(simulation):
    simulation.__str__()


def test_disp_message(simulation, capsys):
    simulation.disp_messages = True
    simulation.disp_message('test message')
    captured = capsys.readouterr()
    assert captured.out == 'test message\n'
    simulation.disp_messages = False
    simulation.disp_message('do not show this')
    captured = capsys.readouterr()
    assert captured.out == ''


def test_simulation_save(simulation, tmp_path_factory):
    test_array = np.r_[1, 2, 3]
    filename = tmp_path_factory.mktemp("data") / "test"
    simulation.save(filename, {'test_array': test_array})
    with pytest.raises(TypeError):
        simulation.save(filename, test_array)


def test_conv_with_function(simulation):
    def handle(x):
        return np.exp(-x**2)

    x = np.r_[0:10:1]
    y = np.r_[0:10:1]
    assert np.allclose(simulation.conv_with_function(y, x, handle),
                       np.array([0.22840624, 1.01047185, 2.00006975, 3.00000006, 4.,
                                 4.99999943, 5.99937279, 6.90631129, 6.02811867, 6.02811867]))


def test_cache_dir(simulation):
    assert simulation.cache_dir == './'



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from udkm1Dsim import Heat
from udkm1Dsim import u
import numpy as np
import pytest
from pint.testing import assert_allclose


# fixtures


@pytest.fixture(scope='module')
def heat(structure, tmp_path_factory):
    return Heat(structure, force_recalc=True, cache_dir=tmp_path_factory.mktemp('cache'),
                save_data=True, disp_messages=True, progress_bar=True,
                )


# tests


def test_heat_str(heat):
    heat.__str__()


def test_get_hash(heat):
    heat.get_hash(np.r_[0:10:1], 300)


def test_check_initial_temperature(heat):
    assert np.allclose(heat.check_initial_temperature(300), 300)
    with pytest.raises(ValueError):
        heat.check_initial_temperature(np.r_[300, 300])


def test_check_excitation(heat):
    heat.heat_diffusion = False
    heat.excitation = {'fluence': [10]*u.mJ/u.cm**2,
                       'delay_pump': [0]*u.ps,
                       'pulse_width': [0]*u.ps,
                       'wavelength': 800*u.nm,
                       'polarization': 'p',
                       'theta': 45*u.deg,
                       'multilayer_absorption': True,
                       'backside': False}
    res, fluence, delay_pump, pulse_width = heat.check_excitation(np.r_[-10:10:1]*u.ps)
    assert np.allclose(fluence, [70.71067812])
    assert np.allclose(delay_pump, [0.])
    assert np.allclose(pulse_width, [0.])

    heat.heat_diffusion = True
    heat.excitation = {'fluence': [10]*u.mJ/u.cm**2,
                       'delay_pump': [0]*u.ps,
                       'pulse_width': [1]*u.ps,
                       'wavelength': 800*u.nm,
                       'polarization': 's',
                       'theta': 45*u.deg,
                       'multilayer_absorption': True,
                       'backside': False}
    res, fluence, delay_pump, pulse_width = heat.check_excitation(np.r_[-10:10:1]*u.ps)
    assert np.allclose(fluence, [70.71067812])
    assert np.allclose(delay_pump, [0.])
    assert np.allclose(pulse_width, [1.0e-12])

    with pytest.raises(Warning):
        heat.excitation = {'fluence': [10]*u.mJ/u.cm**2,
                           'delay_pump': [0]*u.ps,
                           'pulse_width': [1]*u.ps,
                           'wavelength': 800*u.nm,
                           'polarization': 'elliptical',
                           'theta': 45*u.deg,
                           'multilayer_absorption': True,
                           'backside': False}


def test_get_absorption_profile(heat):
    heat.excitation = {'fluence': [10]*u.mJ/u.cm**2,
                       'delay_pump': [0]*u.ps,
                       'pulse_width': [0]*u.ps,
                       'wavelength': 800*u.nm,
                       'polarization': 'p',
                       'theta': 45*u.deg,
                       'multilayer_absorption': False,
                       'backside': False}
    heat.get_absorption_profile()

    heat.excitation = {'fluence': [10]*u.mJ/u.cm**2,
                       'delay_pump': [0]*u.ps,
                       'pulse_width': [0]*u.ps,
                       'wavelength': 800*u.nm,
                       'polarization': 'p',
                       'theta': 45*u.deg,
                       'multilayer_absorption': False,
                       'backside': True}
    heat.get_absorption_profile()

    heat.excitation = {'fluence': [10]*u.mJ/u.cm**2,
                       'delay_pump': [0]*u.ps,
                       'pulse_width': [0]*u.ps,
                       'wavelength': 800*u.nm,
                       'polarization': 'p',
                       'theta': 45*u.deg,
                       'multilayer_absorption': True,
                       'backside': False}
    heat.get_absorption_profile()

    heat.excitation = {'fluence': [10]*u.mJ/u.cm**2,
                       'delay_pump': [0]*u.ps,
                       'pulse_width': [0]*u.ps,
                       'wavelength': 800*u.nm,
                       'polarization': 'p',
                       'theta': 45*u.deg,
                       'multilayer_absorption': True,
                       'backside': True}
    heat.get_absorption_profile()

    heat.excitation = {'fluence': [10]*u.mJ/u.cm**2,
                       'delay_pump': [0]*u.ps,
                       'pulse_width': [0]*u.ps,
                       'wavelength': 800*u.nm,
                       'polarization': 's',
                       'theta': 45*u.deg,
                       'multilayer_absorption': True,
                       'backside': False}
    heat.get_absorption_profile()

    heat.excitation = {'fluence': [10]*u.mJ/u.cm**2,
                       'delay_pump': [0]*u.ps,
                       'pulse_width': [0]*u.ps,
                       'wavelength': 800*u.nm,
                       'polarization': 'p',
                       'theta': 45*u.deg,
                       'multilayer_absorption': False,
                       'backside': False}
    heat.get_absorption_profile(distances=np.r_[0:10e-9:1e-10])


def test_get_temperature_after_delta_excitation(heat):
    heat.excitation = {'fluence': [10]*u.mJ/u.cm**2,
                       'delay_pump': [0]*u.ps,
                       'pulse_width': [0]*u.ps,
                       'wavelength': 800*u.nm,
                       'polarization': 'p',
                       'theta': 45*u.deg,
                       'multilayer_absorption': True,
                       'backside': False}
    heat.get_temperature_after_delta_excitation(10, 300)
    heat.get_temperature_after_delta_excitation(10, 300, distances=np.r_[0:10e-9:1e-10])


def test_get_temp_map(heat):
    print('\n test_get_temp_map\n')
    delays = np.r_[-1:10:0.01]*u.ps
    init_temp = 300*u.K
    heat.backend = 'scipy'
    heat.heat_diffusion = False
    heat.excitation = {'fluence': [10]*u.mJ/u.cm**2,
                       'delay_pump': [0]*u.ps,
                       'pulse_width': [0]*u.ps,
                       'wavelength': 800*u.nm,
                       'polarization': 'p',
                       'theta': 45*u.deg,
                       'multilayer_absorption': True,
                       'backside': False}
    heat.get_temp_map(delays, init_temp)

    # load data from cache
    heat.force_recalc = False
    heat.get_temp_map(delays, init_temp)

    heat.backend = 'matlab'
    heat.force_recalc = True
    heat.heat_diffusion = True
    heat.excitation = {'fluence': [10]*u.mJ/u.cm**2,
                       'delay_pump': [0]*u.ps,
                       'pulse_width': [0.1]*u.ps,
                       'wavelength': 800*u.nm,
                       'polarization': 'p',
                       'theta': 45*u.deg,
                       'multilayer_absorption': True,
                       'backside': False}
    with pytest.raises(Warning):
        heat.get_temp_map(delays, init_temp)

    heat.backend = 'scipy'


@pytest.mark.skip(reason="takes too much time in CI")
def test_get_diffusion_temp_map(heat):
    print('\n test_get_temp_map\n')
    delays = np.r_[-1:10:0.01]*u.ps
    init_temp = 300*u.K
    heat.backend = 'scipy'
    heat.heat_diffusion = True
    heat.excitation = {'fluence': [10]*u.mJ/u.cm**2,
                       'delay_pump': [0]*u.ps,
                       'pulse_width': [0]*u.ps,
                       'wavelength': 800*u.nm,
                       'polarization': 'p',
                       'theta': 45*u.deg,
                       'multilayer_absorption': True,
                       'backside': False}
    heat.get_temp_map(delays, init_temp)


def test_ode_func(heat):
    t = np.r_[0:1]
    u = np.r_[0:2]
    N = 2
    K = 1
    d_x_grid = np.r_[0.5, 0.5]
    x = np.r_[0:2]
    thermal_conds = [[lambda T: 1], [lambda T: 1]]
    heat_capacities = [[lambda T: 1], [lambda T: 1]]
    sub_system_coupling = [[lambda T: 1], [lambda T: 1]]
    densities = np.r_[1:3]
    indices = np.r_[0:2]
    dAdz = np.r_[0:2]
    fluence = np.r_[1]
    delay_pump = np.r_[0]
    pulse_length = np.r_[1]
    bc_top_type = 'isolator'
    bc_top_value = np.r_[1]
    bc_bottom_type = 'isolator'
    bc_bottom_value = np.r_[1]
    pbar = None
    state = (0, 1)

    heat.odefunc(t, u, N, K, d_x_grid, x, thermal_conds, heat_capacities,
                 sub_system_coupling, densities, indices, dAdz, fluence,
                 delay_pump, pulse_length, bc_top_type, bc_top_value,
                 bc_bottom_type, bc_bottom_value, pbar, state)


def test_boundary_conditions(heat):
    heat.boundary_conditions = {'top_type': 'isolator', 'bottom_type': 'isolator'}
    assert heat.boundary_conditions == {'top_type': 'isolator', 'bottom_type': 'isolator'}

    heat.boundary_conditions = {'top_type': 'temperature', 'top_value': 300*u.K,
                                'bottom_type': 'temperature', 'bottom_value': 300*u.K}
    assert heat.boundary_conditions == {'top_type': 'temperature', 'top_value': 300*u.K,
                                        'bottom_type': 'temperature', 'bottom_value': 300*u.K}

    heat.boundary_conditions = {'top_type': 'flux', 'top_value': 1*u.W/u.m**2,
                                'bottom_type': 'flux', 'bottom_value': 1*u.W/u.m**2}
    assert heat.boundary_conditions == {'top_type': 'flux', 'top_value': 1*u.W/u.m**2,
                                        'bottom_type': 'flux', 'bottom_value': 1*u.W/u.m**2}

    with pytest.raises(ValueError):
        heat.boundary_conditions = {'top_type': 'test'}
    with pytest.raises(ValueError):
        heat.boundary_conditions = {'bottom_type': 'test'}
    with pytest.raises(ValueError):
        heat.boundary_conditions = {'top_value': [1, 2]*u.W/u.m**2}
    with pytest.raises(ValueError):
        heat.boundary_conditions = {'bottom_value': [1, 2]*u.W/u.m**2}
    with pytest.raises(ValueError):
        heat.boundary_conditions = 'test'


def test_distances(heat):
    heat.distances = np.r_[0:10e-9:1e-10]*u.nm
    assert_allclose(heat.distances, np.r_[0:10e-9:1e-10]*u.nm)
